#!/usr/bin/python3
########################################################################################
# minigrid.py - Energy-system main module for CLOVER.                                  #
#                                                                                      #
# Authors: Phil Sandwell, Ben Winchester                                               #
# Copyright: Phil Sandwell, 2018                                                       #
# Date created: 13/07/2021                                                             #
# License: Open source                                                                 #
#                                                                                      #
# For more information, please email:                                                  #
#   philip.sandwell@gmail.com                                                          #
########################################################################################
"""
minigrid.py - The energy-system module for CLOVER.

This module carries out a simulation for an energy system based on the various inputs
and profile files that have been parsed/generated.

"""

import datetime
import math

from logging import Logger
from typing import Dict, List, Optional, Tuple, Union

import numpy as np  # pylint: disable=import-error
import pandas as pd  # pylint: disable=import-error

from tqdm import tqdm

from ..__utils__ import (
    AuxiliaryHeaterType,
    BColours,
    CleanWaterMode,
    ColdWaterSupply,
    ColumnHeader,
    DieselMode,
    HOURS_PER_YEAR,
    HTFMode,
    InputFileError,
    InternalError,
    ProgrammerJudgementFault,
    RenewableEnergySource,
    ResourceType,
    Location,
    Scenario,
    Simulation,
    SystemDetails,
    dict_to_dataframe,
)
from ..conversion.conversion import Converter, ThermalDesalinationPlant, WaterSource
from ..generation.solar import solar_degradation
from ..load.load import compute_processed_load_profile, population_hourly
from .__utils__ import determine_available_converters, Minigrid
from .diesel import (
    DieselWaterHeater,
    get_diesel_energy_and_times,
    get_diesel_fuel_usage,
)
from .solar import calculate_pvt_output
from .storage import (
    battery_iteration_step,
    cw_tank_iteration_step,
    get_electric_battery_storage_profile,
    get_water_storage_profile,
)
from .storage_utils import CleanWaterTank

__all__ = (
    "Minigrid",
    "run_simulation",
)


def _calculate_backup_diesel_generator_usage(
    blackout_times: pd.DataFrame,
    minigrid: Minigrid,
    scenario: Scenario,
    total_electric_load: float,
    unmet_energy: pd.DataFrame,
) -> Tuple[float, pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Calculates the backup diesel generator usage based on the scenario.

    Inputs:
        - blackout_times:
            The times for which the system experienced a blackout.
        - minigrid:
            The :class:`Minigrid` being considered.
        - scenario:
            The :class:`Scenario` being used for the run.
        - unmet_energy:
            Load profile of currently unment energy
        - total_electric_load:
            The total electric load placed on the system (kWh).

    Outputs:
        - diesel_capacity:
            The capacity of diesel that needed to be installed to meet the demand.
        - diesel_energy:
            The total energy that was supplied by the diesel system.
        - diesel_fuel_usage:
            The total amount of fuel that was consumed byt he diesel generators.
        - diesel_times:
            The times forw hich the diesel generator was operating.
        - unmet_energy:
            The remaining energy demand which went uynmet after the diesel generator
            fulfilled demand to the :class:`Scenario`'s specification.

    """

    if scenario.diesel_scenario.backup_threshold is None:
        raise InputFileError(
            "diesel inputs",
            "Diesel mode `backup` was selected but no backup threshold was "
            "specified.",
        )
    if minigrid.diesel_generator is None:
        raise InputFileError(
            "energy system inputs",
            "No backup diesel generato was provided on the energy system despite "
            "the diesel mode `backup` being selected.",
        )
    diesel_energy, diesel_times = get_diesel_energy_and_times(
        unmet_energy,
        blackout_times,
        float(scenario.diesel_scenario.backup_threshold),
        total_electric_load,
        scenario.diesel_scenario.mode,
    )
    diesel_capacity: float = float(math.ceil(np.max(diesel_energy, axis=0)))
    diesel_fuel_usage = pd.DataFrame(
        get_diesel_fuel_usage(
            int(diesel_capacity),
            minigrid.diesel_generator,
            diesel_energy,
            diesel_times,
        ).values
    )
    unmet_energy = pd.DataFrame(unmet_energy.values - diesel_energy.values)
    diesel_energy = diesel_energy.abs()  # type: ignore

    return diesel_capacity, diesel_energy, diesel_fuel_usage, diesel_times, unmet_energy


def _calculate_electric_desalination_parameters(
    converters: List[Converter],
    feedwater_sources: List[Converter],
    logger: Logger,
    scenario: Scenario,
) -> Tuple[List[Converter], float, float]:
    """
    Calculates parameters needed for computing electric desalination.

    Inputs:
        - converters:
            The `list` of :class:`Converter` instances defined for the system.
        - feedwater_sources:
            The `list` of :class:`WaterSource` instances that produce feedwater as their
            outputs.
        - logger:
            The :class:`logging.Logger` to use for the run.
        - scenario:
            The :class:`Scenario` for the run.

    Outputs:
        - The `list` of electric desalinators :class:`Converter` instances defined on
          the system.
        - The electric energy consumed per desalinated litre of water produced.
        - The maximum throughput of the electric desalination system.

    """

    # If the mode is backup or prioritise.
    if (
        scenario.desalination_scenario is not None
        and scenario.desalination_scenario.clean_water_scenario.mode
        in {CleanWaterMode.BACKUP, CleanWaterMode.PRIORITISE}
    ):
        # Initialise deslination converters.
        electric_desalinators: List[Converter] = sorted(
            [
                converter
                for converter in converters
                if list(converter.input_resource_consumption)
                == [ResourceType.ELECTRIC, ResourceType.UNCLEAN_WATER]
                and converter.output_resource_type == ResourceType.CLEAN_WATER
            ]
        )

        # Raise an error if there were no electric desalinators defined.
        if len(electric_desalinators) == 0:
            logger.error(
                "%sNo electric desalinators defined despite the desalination mode "
                "being %s%s",
                BColours.fail,
                scenario.desalination_scenario.clean_water_scenario.mode.value,
                BColours.endc,
            )
            raise InputFileError(
                "desalination scenario",
                "No electric desalination devices defined but are required by the "
                "scenario.",
            )
        logger.debug(
            "Electric desalinators: %s",
            ", ".join(str(entry) for entry in electric_desalinators),
        )

        # Compute the amount of energy required per litre desalinated.
        energy_per_desalinated_litre: float = 0.001 * float(
            np.mean(
                [
                    desalinator.input_resource_consumption[ResourceType.ELECTRIC]
                    / desalinator.maximum_output_capacity
                    + desalinator.input_resource_consumption[ResourceType.UNCLEAN_WATER]
                    * feedwater_sources[0].input_resource_consumption[
                        ResourceType.ELECTRIC
                    ]
                    / desalinator.maximum_output_capacity
                    for desalinator in electric_desalinators
                ]
            )
        )

        # Compute the maximum throughput
        maximum_water_throughput: float = min(
            sum(
                desalinator.maximum_output_capacity
                for desalinator in electric_desalinators
            ),
            sum(source.maximum_output_capacity for source in feedwater_sources),
        )
    else:
        electric_desalinators = []
        energy_per_desalinated_litre = 0
        maximum_water_throughput = 0

    return electric_desalinators, energy_per_desalinated_litre, maximum_water_throughput


def _calculate_renewable_cw_profiles(  # pylint: disable=too-many-locals, too-many-statements
    converters: List[Converter],
    disable_tqdm: bool,
    end_hour: int,
    irradiance_data: pd.Series,
    logger: Logger,
    minigrid: Minigrid,
    number_of_cw_tanks: int,
    pvt_size: int,
    scenario: Scenario,
    start_hour: int,
    temperature_data: pd.Series,
    wind_speed_data: Optional[pd.Series],
) -> Tuple[
    Optional[pd.DataFrame],
    pd.DataFrame,
    List[Converter],
    Optional[pd.DataFrame],
    pd.DataFrame,
    pd.DataFrame,
    List[Converter],
    pd.DataFrame,
]:
    """
    Calculates PV-T related profiles.

    Inputs:
        - converters:
            The `list` of :class:`Converter` instances available to be used.
        - disable_tqdm:
            Whether to disable the tqdm progress bars (True) or display them (False).
        - end_hour:
            The final hour for which the simulation will be carried out.
        - irradiance_data:
            The total solar irradiance data.
        - logger:
            The :class:`logging.Logger` to use for the run.
        - minigrid:
            The energy system being considered.
        - number_of_cw_tanks:
            The number of clean-water tanks installed within the system.
        - pvt_size:
            Amount of PV-T in PV-T units.
        - scenario:
            The scenario being considered.
        - start_hour:
            The first hour for which the simulation will be carried out.
        - temperature_data:
            The temperature data series.
        - wind_speed_data:
            The wind-speed data series.

    Outputs:
        - buffer_tank_temperature:
            The temperature of the buffer tank, measured in degrees Celcius.
        - buffer_tank_volume_supplied:
            The volume of buffer solution outputted by the HTF buffer tanks.
        - feedwater_sources:
            The :class:`Converter` instances which are a source of feedwater to the PV-T
            system.
        - clean_water_pvt_collector_output_temperature:
            The output temperature of HTF from the PV-T collectors, measured in degrees
            Celcius.
        - clean_water_pvt_electric_power_per_unit:
            The electric power produced by the PV-T, in kWh, per unit of PV-T installed.
        - renewable_cw_produced:
            The amount of clean water produced renewably, measured in litres.
        - required_feedwater_sources:
            The `list` of feedwater sources required to supply the needs of the
            desalination system.
        - thermal_desalination_electric_power_consumed:
            The electric power consumed in operating the thermal desalination plant,
            measured in kWh.

    """

    if scenario.desalination_scenario is not None:
        # Determine the list of available feedwater sources.
        logger.info("Determining available feedwater sources.")
        feedwater_sources: List[Converter] = sorted(
            [
                converter
                for converter in converters
                if list(converter.input_resource_consumption) == [ResourceType.ELECTRIC]
                and converter.output_resource_type == ResourceType.UNCLEAN_WATER
            ]
        )
        logger.debug(
            "Available feedwater sources determined: %s",
            ", ".join([str(source) for source in feedwater_sources])
            if len(feedwater_sources) > 0
            else "",
        )
    else:
        feedwater_sources = []

    if scenario.pv_t and scenario.desalination_scenario is not None:
        logger.info("Calculating clean-water PV-T performance profiles.")
        if wind_speed_data is None:
            raise InternalError(
                "Wind speed data required in PV-T computation and not passed to the "
                "energy system module."
            )
        if minigrid.water_pump is None:
            logger.error(
                "%sNo water pump defined on the minigrid despite PV-T modelling being "
                "requested via the scenario files.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InternalError(
                "No water pump defined as part of the energy system despite the PV-T "
                "modelling being requested."
            )

        # Determine the thermal desalination plant being used.
        logger.info("Determining desalination plant.")
        try:
            thermal_desalination_plant: ThermalDesalinationPlant = [
                converter
                for converter in converters
                if isinstance(converter, ThermalDesalinationPlant)
            ][0]
        except IndexError:
            logger.error(
                "%sNo valid thermal desalination plants specified despite PV-T being "
                "specified.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InputFileError(
                "conversion inputs", "No valid thermal desalination plants specified."
            ) from None
        logger.info(
            "Desalination plant determined: %s", thermal_desalination_plant.name
        )

        # Determine whether the water pump is capable for supplying the PV-T panels with
        # enough throughput.
        if (
            scenario.desalination_scenario.pvt_scenario.mass_flow_rate * pvt_size
            > minigrid.water_pump.throughput
        ):
            logger.error(
                "%sThe water pump supplied, %s, is incapable of meeting the required "
                "PV-T flow rate of %s litres/hour. Max pump throughput: %s litres/hour."
                "%s",
                BColours.fail,
                minigrid.water_pump.name,
                scenario.desalination_scenario.pvt_scenario.mass_flow_rate * pvt_size,
                minigrid.water_pump.throughput,
                BColours.endc,
            )
            raise InputFileError(
                "transmission inputs",
                "The water pump defined is unable to meet PV-T flow requirements.",
            )

        if thermal_desalination_plant.htf_mode == HTFMode.CLOSED_HTF:
            thermal_desalination_plant_input_type: ResourceType = (
                ResourceType.UNCLEAN_WATER
            )
        if thermal_desalination_plant.htf_mode == HTFMode.FEEDWATER_HEATING:
            thermal_desalination_plant_input_type = ResourceType.HOT_UNCLEAN_WATER
        if thermal_desalination_plant.htf_mode == HTFMode.COLD_WATER_HEATING:
            logger.error(
                "%sCold-water heating thermal desalination plants are not supported.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InputFileError(
                "converter inputs OR desalination scenario",
                f"The htf mode '{HTFMode.COLD_WATER_HEATING.value}' is not currently "
                "supported.",
            )

        thermal_desalination_plant_input_flow_rate = (
            thermal_desalination_plant.input_resource_consumption[
                thermal_desalination_plant_input_type
            ]
        )

        if (
            sum(
                feedwater_source.maximum_output_capacity
                for feedwater_source in feedwater_sources
            )
            < thermal_desalination_plant_input_flow_rate
        ):
            logger.error(
                "%sThe feedwater sources are unable to supply enough throughput to "
                "facilitate the thermal desalination plant. If you are running a "
                "simulation, consider using a smaller desalination plant or a larger "
                "number of feedwater sources. If you are running an optimisation, "
                "consider using a greater number of feedwater sources as your initial "
                "maximum point. Or, it is possible that no feedwater sources have been "
                "defined within your optimisation inputs file.%s",
                BColours.fail,
                BColours.endc,
            )
            logger.info(
                "Feedwater sources: %s",
                ", ".join([str(source) for source in feedwater_sources]),
            )
            logger.info("Desalination plant: %s", thermal_desalination_plant)
            raise InputFileError(
                "desalination scenario",
                "The feedwater sources cannot meet the thermal desalination plant "
                "input demand.",
            )

        logger.info("Determining required feedwater sources.")
        feedwater_capacity: float = 0
        required_feedwater_sources: List[Converter] = []
        while (
            feedwater_capacity
            < thermal_desalination_plant.input_resource_consumption[
                thermal_desalination_plant_input_type
            ]
        ):
            required_feedwater_sources.append(feedwater_sources.pop(0))
            feedwater_capacity += required_feedwater_sources[-1].maximum_output_capacity

        feedwater_sources.extend(required_feedwater_sources)
        logger.info("Required feedwater sources determined.")
        logger.debug(
            "Required feedwater sources: %s",
            ", ".join([str(source) for source in required_feedwater_sources]),
        )

        # Compute the output of the PV-T system.
        clean_water_pvt_collector_output_temperature: Optional[pd.DataFrame]
        buffer_tank_temperature: Optional[pd.DataFrame]
        (
            clean_water_pvt_collector_output_temperature,
            clean_water_pvt_electric_power_per_unit,
            clean_water_pvt_pump_times,
            buffer_tank_temperature,
            buffer_tank_volume_supplied,
        ) = calculate_pvt_output(
            disable_tqdm,
            end_hour,
            irradiance_data[start_hour:end_hour],
            logger,
            minigrid,
            number_of_cw_tanks,
            None,
            pvt_size,
            ResourceType.CLEAN_WATER,
            scenario,
            start_hour,
            temperature_data[start_hour:end_hour],
            thermal_desalination_plant,
            wind_speed_data[start_hour:end_hour],
        )
        logger.info("PV-T performance successfully computed.")

        # Compute the clean water supplied by the desalination unit.
        renewable_cw_produced: pd.DataFrame = (
            buffer_tank_volume_supplied > 0
        ) * thermal_desalination_plant.maximum_output_capacity

        # Compute the power consumed by the thermal desalination plant.
        thermal_desalination_electric_power_consumed: pd.DataFrame = pd.DataFrame(
            (
                (renewable_cw_produced > 0)
                * (
                    0.001
                    * thermal_desalination_plant.input_resource_consumption[
                        ResourceType.ELECTRIC
                    ]
                    + 0.001
                    * sum(
                        source.input_resource_consumption[ResourceType.ELECTRIC]
                        for source in required_feedwater_sources
                    )
                )
            ).values
            + (clean_water_pvt_pump_times > 0) * 0.001 * minigrid.water_pump.consumption
        )

        buffer_tank_temperature = buffer_tank_temperature.reset_index(drop=True)
        clean_water_pvt_collector_output_temperature = (
            clean_water_pvt_collector_output_temperature.reset_index(drop=True)
        )
        clean_water_pvt_electric_power_per_unit = (
            clean_water_pvt_electric_power_per_unit.reset_index(drop=True)
        )
        renewable_cw_produced = renewable_cw_produced.reset_index(drop=True)
        buffer_tank_volume_supplied = buffer_tank_volume_supplied.reset_index(drop=True)
        thermal_desalination_electric_power_consumed = (
            thermal_desalination_electric_power_consumed.reset_index(drop=True)
        )
        logger.info("Clean-water PV-T performance profiles determined.")

    else:
        logger.info("Skipping clean-water PV-T performance-profile calculation.")
        buffer_tank_temperature = None
        buffer_tank_volume_supplied = pd.DataFrame([0] * (end_hour - start_hour))
        clean_water_pvt_collector_output_temperature = None
        clean_water_pvt_electric_power_per_unit = pd.DataFrame(
            [0] * (end_hour - start_hour)
        )
        renewable_cw_produced = pd.DataFrame([0] * (end_hour - start_hour))
        required_feedwater_sources = []
        thermal_desalination_electric_power_consumed = pd.DataFrame(
            [0] * (end_hour - start_hour)
        )

    return (
        buffer_tank_temperature,
        buffer_tank_volume_supplied,
        feedwater_sources,
        clean_water_pvt_collector_output_temperature,
        clean_water_pvt_electric_power_per_unit,
        renewable_cw_produced,
        required_feedwater_sources,
        thermal_desalination_electric_power_consumed,
    )


def _calculate_renewable_hw_profiles(  # pylint: disable=too-many-locals, too-many-statements
    converters: List[Converter],
    disable_tqdm: bool,
    end_hour: int,
    irradiance_data: pd.Series,
    logger: Logger,
    minigrid: Minigrid,
    number_of_hw_tanks: int,
    processed_total_hw_load: pd.DataFrame,
    pvt_size: int,
    scenario: Scenario,
    start_hour: int,
    temperature_data: pd.Series,
    wind_speed_data: Optional[pd.Series],
) -> Tuple[
    Optional[Union[Converter, DieselWaterHeater]],
    pd.DataFrame,
    Optional[pd.DataFrame],
    pd.DataFrame,
    Optional[pd.DataFrame],
    Optional[pd.DataFrame],
    Optional[pd.DataFrame],
]:
    """
    Calculates PV-T related profiles for the hot-water system.

    Inputs:
        - converters:
            The `list` of :class:`Converter` instances available to be used.
        - disable_tqdm:
            Whether to disable the tqdm progress bars (True) or display them (False).
        - end_hour:
            The final hour for which the simulation will be carried out.
        - irradiance_data:
            The total solar irradiance data.
        - logger:
            The :class:`logging.Logger` to use for the run.
        - minigrid:
            The energy system being considered.
        - number_of_hw_tanks:
            The number of hot-water tanks installed.
        - processed_total_hw_load:
            The total hot-water load placed on the system, defined in litres/hour at
            every time step.
        - pvt_size:
            Amount of PV-T in PV-T units.
        - scenario:
            The scenario being considered.
        - start_hour:
            The first hour for which the simulation will be carried out.
        - temperature_data:
            The temperature data series.
        - wind_speed_data:
            The wind-speed data series.

    Outputs:
        - auxiliary_heater:
            The auxiliary heater associated with the system.
        - hot_water_power_consumed:
            The electric power consumed by the hot-water system, including any water
            pumps and electricity that was used meeting unmet hot-water demand.
        - hot_water_pvt_collector_output_temperature:
            The output temperature from the PV-T panels associated with the hot-water
            system.
        - hot_water_pvt_electric_power_per_unit:
            The electric power produced by the PV-T, in kWh, per unit of PV-T installed.
        - hot_water_tank_temperature:
            The temperature of the hot-water tank, in degrees Celcius, at each time
            step throughout the simulation period.
        - hot_water_tank_volume_supplied:
            The volume of hot-water supplied by the hot-water tank.
        - renewable_hw_fraction:
            The fraction of the hot-water demand which was covered using renewables vs
            which was covered using auxiliary means.

    """

    if scenario.pv_t and scenario.hot_water_scenario is not None:
        logger.info("Calculating hot-water PV-T performance profiles.")
        if wind_speed_data is None:
            raise InternalError(
                "Wind speed data required in PV-T computation and not passed to the "
                "energy system module."
            )

        if scenario.hot_water_scenario.cold_water_supply != ColdWaterSupply.UNLIMITED:
            logger.error(
                "%sOnly '%s' cold-water supplies for the hot-water system are "
                "currently supported.%s",
                BColours.fail,
                ColdWaterSupply.UNLIMITED.value,
                BColours.endc,
            )

        if minigrid.hot_water_tank is None:
            logger.error(
                "%sNo hot-water tank was defined for the minigrid despite hot-water"
                "modelling being requested.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InternalError(
                "No water pump defined as part of the energy system despite the PV-T "
                "modelling being requested."
            )

        if minigrid.water_pump is None:
            logger.error(
                "%sNo water pump defined on the minigrid despite PV-T modelling being "
                "requested via the scenario files.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InternalError(
                "No water pump defined as part of the energy system despite the PV-T "
                "modelling being requested."
            )

        # Determine whether the water pump is capable for supplying the PV-T panels with
        # enough throughput.
        if (
            scenario.hot_water_scenario.pvt_scenario.mass_flow_rate * pvt_size
            > minigrid.water_pump.throughput
        ):
            logger.error(
                "%sThe water pump supplied, %s, is incapable of meeting the required "
                "PV-T flow rate of %s litres/hour. Max pump throughput: %s litres/hour."
                "%s",
                BColours.fail,
                minigrid.water_pump.name,
                scenario.hot_water_scenario.pvt_scenario.mass_flow_rate * pvt_size,
                minigrid.water_pump.throughput,
                BColours.endc,
            )
            raise InputFileError(
                "transmission inputs",
                "The water pump defined is unable to meet PV-T flow requirements.",
            )

        # Determine the auxiliary heater associated with the system.
        if scenario.hot_water_scenario.auxiliary_heater == AuxiliaryHeaterType.DIESEL:
            auxiliary_heater: Optional[
                Union[Converter, DieselWaterHeater]
            ] = minigrid.diesel_water_heater
        if scenario.hot_water_scenario.auxiliary_heater == AuxiliaryHeaterType.ELECTRIC:
            try:
                auxiliary_heater = [
                    converter
                    for converter in converters
                    if converter.output_resource_type == ResourceType.HOT_CLEAN_WATER
                    and ResourceType.ELECTRIC in converter.input_resource_consumption
                    and ResourceType.CLEAN_WATER in converter.input_resource_consumption
                ][0]
            except IndexError:
                logger.error(
                    "%sFailed to determine electric water heater despite an electric "
                    "auxiliary hot-water type being selected.%s",
                    BColours.fail,
                    BColours.endc,
                )
                raise InputFileError(
                    "energy system inputs OR hot-water scenario",
                    "No electric water heater defined despite the hot-water scenario "
                    "specifying that this is needed.",
                ) from None

        if auxiliary_heater is None:
            logger.error(
                "%sDiesel water heater not defined despite hot-water auxiliary "
                "heating mode being specified as diesel.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InputFileError(
                "energy system inputs OR hot-water scenario",
                "No diesel hot-water heater defined despite the hot-water "
                "scenario specifying that this is needed.",
            )

        logger.info("Auxiliary heater successfully determined.")
        logger.debug("Auxiliary heater: %s", str(auxiliary_heater))

        # Compute the output of the PV-T system.
        hot_water_pvt_collector_output_temperature: Optional[pd.DataFrame]
        hot_water_pvt_electric_power_per_unit: pd.DataFrame
        hot_water_pvt_pump_times: pd.DataFrame
        hot_water_tank_temperature: Optional[pd.DataFrame]
        hot_water_tank_volume_supplied: Optional[pd.DataFrame]
        (
            hot_water_pvt_collector_output_temperature,
            hot_water_pvt_electric_power_per_unit,
            hot_water_pvt_pump_times,
            hot_water_tank_temperature,
            hot_water_tank_volume_supplied,
        ) = calculate_pvt_output(
            disable_tqdm,
            end_hour,
            irradiance_data[start_hour:end_hour],
            logger,
            minigrid,
            number_of_hw_tanks,
            processed_total_hw_load.iloc[:, 0],
            pvt_size,
            ResourceType.HOT_CLEAN_WATER,
            scenario,
            start_hour,
            temperature_data[start_hour:end_hour],
            None,
            wind_speed_data[start_hour:end_hour],
        )
        logger.info("Hot-water PV-T performance successfully computed.")

        # Compute the heat consumed by the auxiliary heater.
        # auxiliary_heater_heat_consumption: pd.DataFrame = pd.DataFrame(
        #     (hot_water_tank_volume_supplied > 0)
        #     * hot_water_tank_volume_supplied  # type: ignore
        #     * minigrid.hot_water_tank.heat_capacity
        #     * (
        #         scenario.hot_water_scenario.demand_temperature  # type: ignore
        #         - hot_water_tank_temperature
        #     )
        # )

        # Compute the electric power consumed by the auxiliary heater.
        auxiliary_heater_power_consumption: pd.DataFrame = pd.DataFrame(
            0.001
            * auxiliary_heater.input_resource_consumption[
                ResourceType.ELECTRIC
            ]  # [Wh/degC]
            * (
                hot_water_tank_volume_supplied  # type: ignore
                / auxiliary_heater.input_resource_consumption[ResourceType.CLEAN_WATER]
            )  # [operating fraction]
            * (hot_water_tank_volume_supplied > 0)  # type: ignore
            * (  # type: ignore
                scenario.hot_water_scenario.demand_temperature  # type: ignore
                - hot_water_tank_temperature
            )
        )

        # Compute the power consumed by the thermal desalination plant.
        hot_water_power_consumed: pd.DataFrame = pd.DataFrame(
            auxiliary_heater_power_consumption
            + 0.001
            * (hot_water_pvt_pump_times > 0)  # type: ignore
            * minigrid.water_pump.consumption
        )

        # Determine the fraction of the output which was met renewably.
        renewable_hw_fraction: Optional[pd.DataFrame] = (  # type: ignore
            hot_water_tank_temperature
            - scenario.hot_water_scenario.cold_water_supply_temperature
        ) / (
            scenario.hot_water_scenario.demand_temperature
            - scenario.hot_water_scenario.cold_water_supply_temperature
        )

        hot_water_power_consumed = hot_water_power_consumed.reset_index(drop=True)
        hot_water_pvt_collector_output_temperature = (
            hot_water_pvt_collector_output_temperature.reset_index(drop=True)
        )
        hot_water_pvt_electric_power_per_unit = (
            hot_water_pvt_electric_power_per_unit.reset_index(drop=True)
        )
        hot_water_tank_temperature = hot_water_tank_temperature.reset_index(drop=True)
        hot_water_tank_volume_supplied = hot_water_tank_volume_supplied.reset_index(
            drop=True
        )
        renewable_hw_fraction = renewable_hw_fraction.reset_index(  # type: ignore
            drop=True
        )
        logger.info("Hot-water PV-T performance profiles determined.")

    else:
        logger.info("Skipping hot-water PV-T performance-profile calculation.")
        auxiliary_heater = None
        hot_water_power_consumed = pd.DataFrame([0] * (end_hour - start_hour))
        hot_water_pvt_collector_output_temperature = None
        hot_water_pvt_electric_power_per_unit = pd.DataFrame(
            [0] * (end_hour - start_hour)
        )
        hot_water_tank_temperature = None
        hot_water_tank_volume_supplied = None
        renewable_hw_fraction = None

    return (
        auxiliary_heater,
        hot_water_power_consumed,
        hot_water_pvt_collector_output_temperature,
        hot_water_pvt_electric_power_per_unit,
        hot_water_tank_temperature,
        hot_water_tank_volume_supplied,
        renewable_hw_fraction,
    )


def _setup_tank_storage_profiles(
    logger: Logger,
    number_of_tanks: int,
    power_consumed: pd.DataFrame,
    resource_type: ResourceType,
    scenario: Scenario,
    tank: Optional[CleanWaterTank],
) -> Tuple[Dict[int, float], float, float, float, Dict[int, float]]:
    """
    Sets up tank storage parameters.

    Inputs:
        - logger:
            The :class:`logging.Logger` to use for the run.
        - number_of_tanks:
            The number of tanks of this type to use for the run.
        - power_consumed:
            The electric power consumed associated with the storage of these
            :class:`ResourceType` tanks.
        - resource_type:
            The :class:`ResourceType` held within the :class:`CleanWaterTank`.
        - scenario:
            The :class:`Scenario` for the run.
        - tank:
            The :class:`CleanWaterTank`, representing either a clean- or hot-water tank,
            to use for the run.

    Outputs:
        - hourly_tank_storage:
            The hourly tank storage.
        - initial_tank_storage:
            The amount of water initially in the tank.
        - minimum_tank_storage:
            The minimum level of the tank permitted.
        - power_consumed_mapping:
            A mapping between time as `int` and the electric power consumed.

    """

    power_consumed_mapping: Dict[int, float] = power_consumed[  # type: ignore
        0
    ].to_dict()

    if (
        resource_type in scenario.resource_types
        and scenario.desalination_scenario is not None
    ):
        if tank is None:
            logger.error(
                "%sNo tank specifeid when attempting to compute %s loads.%s",
                BColours.fail,
                resource_type.value,
                BColours.endc,
            )
            raise InternalError(
                f"No {resource_type.value} tank specified on the energy system despite "
                + f"{resource_type.value} loads being requested.",
            )
        hourly_tank_storage: Dict[int, float] = {}
        initial_tank_storage: float = 0.0

        # Determine the maximum tank storage.
        try:
            maximum_tank_storage: float = (
                number_of_tanks * tank.mass * tank.maximum_charge
            )
        except AttributeError:
            logger.error(
                "%sNo %s water tank provided on the energy system despite associated "
                "demands expected.%s",
                BColours.fail,
                resource_type.value,
                BColours.endc,
            )
            raise InputFileError(
                "energy system OR tank",
                f"No {resource_type.value} water tank was provided on the energy "
                + f"system despite {resource_type.value}-water demands being expected.",
            ) from None

        try:
            minimum_tank_storage: float = (
                number_of_tanks * tank.mass * tank.minimum_charge
            )
        except AttributeError:
            logger.error(
                "%sNo %s water tank provided on the energy system despite associated "
                "demands expected.%s",
                BColours.fail,
                resource_type.value,
                BColours.endc,
            )
            raise InputFileError(
                "energy system OR tank",
                f"No {resource_type.value} water tank was provided on the energy "
                + f"system despite {resource_type.value}-water demands being expected.",
            ) from None

    else:
        hourly_tank_storage = {}
        initial_tank_storage = 0
        maximum_tank_storage = 0
        minimum_tank_storage = 0

    return (
        hourly_tank_storage,
        initial_tank_storage,
        maximum_tank_storage,
        minimum_tank_storage,
        power_consumed_mapping,
    )


def _update_battery_health(
    battery_energy_flow: float,
    battery_health: Dict[int, float],
    cumulative_battery_storage_power: float,
    electric_storage_size: float,
    hourly_battery_storage: Dict[int, float],
    maximum_battery_energy_throughput: float,
    minigrid: Minigrid,
    storage_power_supplied: Dict[int, float],
    *,
    time_index: int,
) -> Tuple[float, float, float]:
    """
    Updates the health of the batteries.

    Inputs:
        - battery_energy_flow:
            The net energy flow, into, or out of, the battery.
        - battery_health:
            The battery health at each time step.
        - cumulative_battery_storage_power: float:
            The cumulative amount of power that has been stored in the batteries,
            measured in kWh.
        - electric_storage_size:
            The size of the electric storage system.
        - hourly_battery_storage:
            The battery storage at each time step.
        - maximum_battery_energy_throughput:
            The maximum energy throughput through the batteries.
        - minigrid:
            The :class:`Minigrid` being modelled.
        - storage_power_supplied:
            The amount of power supplied by the storage system.
        - time_index:
            The current time (hour) being considered.

    Outputs:
        - cumulative_battery_storage_power:
            The cumulative amount of electricity that has been stored in the batteries.
        - maximum_battery_storage:
            The newly calculated maximum amount of energy that can be stored in the
            batteries having acounted for battery degredation.
        - minimum_battery_storage:
            The newly calculated minimum amount of energy that can be stored in the
            batteries having acounted for battery degredation.

    """

    if time_index == 0:
        storage_power_supplied[time_index] = 0.0 - battery_energy_flow
    else:
        storage_power_supplied[time_index] = max(
            hourly_battery_storage[time_index - 1]
            * (1.0 - minigrid.battery.leakage)  # type: ignore
            - hourly_battery_storage[time_index],
            0.0,
        )
    cumulative_battery_storage_power += storage_power_supplied[time_index]

    battery_storage_degradation = (
        1.0
        - minigrid.battery.lifetime_loss  # type: ignore
        * (cumulative_battery_storage_power / maximum_battery_energy_throughput)
    )
    maximum_battery_storage = (
        battery_storage_degradation
        * electric_storage_size
        * minigrid.battery.maximum_charge  # type: ignore
        * minigrid.battery.storage_unit  # type: ignore
    )
    minimum_battery_storage = (
        battery_storage_degradation
        * electric_storage_size
        * minigrid.battery.minimum_charge  # type: ignore
        * minigrid.battery.storage_unit  # type: ignore
    )
    battery_health[time_index] = battery_storage_degradation

    return (
        cumulative_battery_storage_power,
        maximum_battery_storage,
        minimum_battery_storage,
    )


def run_simulation(  # pylint: disable=too-many-locals, too-many-statements
    clean_water_pvt_size: int,
    conventional_cw_source_profiles: Optional[Dict[WaterSource, pd.DataFrame]],
    converters: Union[Dict[str, Converter], List[Converter]],
    disable_tqdm: bool,
    electric_storage_size: float,
    grid_profile: Optional[pd.DataFrame],
    hot_water_pvt_size: int,
    irradiance_data: pd.Series,
    kerosene_usage: pd.DataFrame,
    location: Location,
    logger: Logger,
    minigrid: Minigrid,
    number_of_cw_tanks: int,
    number_of_hw_tanks: int,
    pv_power_produced: pd.Series,
    pv_size: float,
    scenario: Scenario,
    simulation: Simulation,
    temperature_data: pd.Series,
    total_loads: Dict[ResourceType, Optional[pd.DataFrame]],
    wind_speed_data: Optional[pd.Series],
) -> Tuple[datetime.timedelta, pd.DataFrame, SystemDetails]:
    """
    Simulates a minigrid system

    This function simulates the energy system of a given capacity and to the parameters
    stated in the input files.

    Inputs:
        - clean_water_pvt_size:
            Amount of PV-T in PV-T units associated with the clean-water system.
        - conventional_cw_source_profiles:
            A mapping between :class:`WaterSource` instances and the associated water
            that can be drawn from the source throughout the duration of the simulation.
        - converters:
            The `list` of :class:`Converter` instances available to be used.
        - disable_tqdm:
            Whether to disable the tqdm progress bars (True) or display them (False).
        - electric_storage_size:
            Amount of storage in terms of the number of batteries included.
        - grid_profile:
            The grid-availability profile.
        - hot_water_pvt_size:
            Amount of PV-T in PV-T units associated with the hot-water system.
        - irradiance_data:
            The total solar irradiance data.
        - kerosene_usage:
            The kerosene-usage profile.
        - location:
            The location being considered.
        - logger:
            The :class:`logging.Logger` to use for the run.
        - minigrid:
            The energy system being considered.
        - number_of_cw_tanks:
            The number of clean-water tanks installed in the system.
        - number_of_hw_tanks:
            The number of hot-water tanks installed in the system.
        - pv_size:
            Amount of PV in PV units.
        - pv_power_produced:
            The total energy outputted by the solar system per PV unit.
        - renewable_cw_produced:
            The amount of clean-water produced renewably, mesaured in litres.
        - scenario:
            The scenario being considered.
        - simulation:
            The simulation to run.
        - temperature_data:
            The temperature data series.
        - total_loads:
            A mapping between :class:`ResourceType`s and their associated total loads
            placed on the system.
        - wind_speed_data:
            The wind-speed data series.

    Outputs:
        - The time taken for the simulation.
        - System performance outputs:
            - system_performance_outputs:
                Hourly performance of the simulated system
            - load_energy:
                Amount of energy (kWh) required to satisfy the loads
            - total_energy_used:
                Amount of energy (kWh) used by the system
            - unmet_energy:
                Amount of energy (kWh) unmet by the system
            - blackout_times:
                Times with power is available (0) or unavailable (1)
            - renewables_energy_used_directly:
                Amount of energy (kWh) from renewables used directly to satisfy load
                (kWh)
            - storage_power_supplied:
                Amount of energy (kWh) supplied by battery storage
            - grid_energy:
                Amount of energy (kWh) supplied by the grid
            - diesel_energy:
                Amount of energy (kWh) supplied from diesel generator
            - diesel_times:
                Times when diesel generator is on (1) or off (0)
            - diesel_fuel_usage:
                Amount of diesel (l) used by the generator
            - battery_storage_profile:
                Amount of energy (kWh) into (+ve) and out of (-ve) the battery
            - renewables_energy:
                Amount of energy (kWh) provided by renewables to the system
            - hourly_battery_storage:
                Amount of energy (kWh) in the battery
            - energy_surplus:
                Amount of energy (kWh) dumped owing to overgeneration
            - battery_health:
                Relative capactiy of the battery compared to new (0.0-1.0)
            - households:
                Number of households in the community
            - kerosene_usage:
                Number of kerosene lamps in use (if no power available)
            - kerosene_mitigation:
                Number of kerosene lamps not used (when power is available)
        - System details about the run.

    """

    # Currently, only systems including batteries are supported.
    if minigrid.battery is None:
        logger.error(
            "%sNo battery information available when calling the energy system.%s",
            BColours.fail,
            BColours.endc,
        )
        raise Exception(
            "No battery information available when calling the energy system."
        )

    # Start timer to see how long simulation will take
    timer_start = datetime.datetime.now()

    # Initialise simulation parameters
    start_hour = simulation.start_year * HOURS_PER_YEAR
    end_hour = simulation.end_year * HOURS_PER_YEAR
    simulation_hours = end_hour - start_hour

    if isinstance(converters, dict):
        available_converters = determine_available_converters(
            converters, logger, minigrid, scenario
        )
        logger.info("Subset of available converters determined.")
    elif isinstance(converters, list):
        available_converters = converters
        logger.info("Converter sizes manually passed in, using information provided.")
    else:
        logger.error(
            "Unsupported type '%s' for converters in call to run_simulation.%s",
            BColours.fail,
            str(type(converters)),
            BColours.endc,
        )
        raise ProgrammerJudgementFault(
            "run-simulation function",
            "Misuse of `converters` parameter when calling `run_simulation`.",
        )
    logger.debug(
        "Available converters: %s",
        ", ".join([str(entry) for entry in available_converters]),
    )
    grid_profile = (
        grid_profile
        if grid_profile is not None
        else pd.DataFrame([0] * simulation_hours)
    )
    total_cw_load: Optional[pd.DataFrame] = total_loads[ResourceType.CLEAN_WATER]
    total_electric_load: Optional[pd.DataFrame] = total_loads[ResourceType.ELECTRIC]
    total_hw_load: Optional[pd.DataFrame] = total_loads[ResourceType.HOT_CLEAN_WATER]

    # Calculate PV-T related performance profiles.
    buffer_tank_temperature: Optional[pd.DataFrame]
    feedwater_sources: List[Converter]
    clean_water_pvt_collector_output_temperature: Optional[pd.DataFrame]
    clean_water_pvt_electric_power_per_unit: pd.DataFrame
    renewable_cw_produced: pd.DataFrame
    buffer_tank_volume_supplied: pd.DataFrame
    thermal_desalination_electric_power_consumed: pd.DataFrame

    (
        buffer_tank_temperature,
        buffer_tank_volume_supplied,
        feedwater_sources,
        clean_water_pvt_collector_output_temperature,
        clean_water_pvt_electric_power_per_unit,
        renewable_cw_produced,
        required_cw_feedwater_sources,
        thermal_desalination_electric_power_consumed,
    ) = _calculate_renewable_cw_profiles(
        available_converters,
        disable_tqdm,
        end_hour,
        irradiance_data,
        logger,
        minigrid,
        number_of_cw_tanks,
        clean_water_pvt_size,
        scenario,
        start_hour,
        temperature_data,
        wind_speed_data,
    )
    logger.debug(
        "Mean buffer tank temperature: %s",
        np.mean(buffer_tank_temperature.values)
        if buffer_tank_temperature is not None
        else "N/A",
    )
    logger.debug(
        "Soruces of feedwater: %s",
        ", ".join([str(source) for source in feedwater_sources])
        if len(feedwater_sources) > 0
        else "N/A",
    )
    logger.debug(
        "Mean clean-water PV-T electric power per unit: %s",
        np.mean(clean_water_pvt_electric_power_per_unit.values),
    )
    logger.debug(
        "Maximum thermal desalination plant power consumption: %s",
        np.max(thermal_desalination_electric_power_consumed.values),
    )
    logger.debug(
        "Mean thermal desalination plant power consumption: %s",
        np.mean(thermal_desalination_electric_power_consumed.values),
    )

    # Calculate clean-water-related performance profiles.
    clean_water_power_consumed: pd.DataFrame
    renewable_cw_used_directly: pd.DataFrame
    tank_storage_profile: pd.DataFrame
    total_cw_supplied: Optional[pd.DataFrame] = None

    if scenario.desalination_scenario is not None:
        if total_cw_load is None:
            raise Exception(
                f"{BColours.fail}A simulation was run that specified a clean-water "
                + f"load but no clean-water load was passed in.{BColours.endc}"
            )
        # Process the load profile based on the relevant scenario.
        processed_total_cw_load: Optional[pd.DataFrame] = pd.DataFrame(
            compute_processed_load_profile(scenario, total_cw_load)[  # type: ignore
                start_hour:end_hour
            ].values
        )

        if processed_total_cw_load is None:
            logger.error(
                "%sNo processed clean-water load was calculated.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InternalError(
                "Failed to calculate the processed total clean-water load profile."
            )

        # Determine the water-tank storage profile.
        (
            clean_water_power_consumed,
            renewable_cw_used_directly,
            tank_storage_profile,
        ) = get_water_storage_profile(
            processed_total_cw_load,
            renewable_cw_produced,
        )
        number_of_buffer_tanks: int = 1
    else:
        clean_water_power_consumed = pd.DataFrame([0] * simulation_hours)
        number_of_buffer_tanks = 0
        processed_total_cw_load = None
        renewable_cw_used_directly = pd.DataFrame([0] * simulation_hours)
        tank_storage_profile = pd.DataFrame([0] * simulation_hours)

    # Calculate hot-water-related profiles.
    processed_total_hw_load: pd.DataFrame
    if scenario.hot_water_scenario is not None:
        if total_hw_load is None:
            raise Exception(
                f"{BColours.fail}A simulation was run that specified a hot-water load "
                + f"but no hot-water load was passed in.{BColours.endc}"
            )
        # Process the load profile based on the relevant scenario.
        processed_total_hw_load = pd.DataFrame(
            compute_processed_load_profile(scenario, total_hw_load)[  # type: ignore
                start_hour:end_hour
            ]
        )
    else:
        number_of_hw_tanks = 0
        processed_total_hw_load = pd.DataFrame([0] * (end_hour - start_hour))

    # Calculate hot-water PV-T related performance profiles.
    hot_water_pump_electric_power_consumed: pd.DataFrame  # pylint: disable=unused-variable
    hot_water_pvt_collector_output_temperature: pd.DataFrame
    hot_water_pvt_electric_power_per_unit: pd.DataFrame
    hot_water_tank_temperature: pd.DataFrame
    hot_water_tank_volume_supplied: pd.DataFrame
    renewable_hw_fraction: pd.DataFrame

    (
        auxiliary_heater,  # pylint: disable=unused-variable
        hot_water_power_consumed,
        hot_water_pvt_collector_output_temperature,
        hot_water_pvt_electric_power_per_unit,
        hot_water_tank_temperature,
        hot_water_tank_volume_supplied,
        renewable_hw_fraction,
    ) = _calculate_renewable_hw_profiles(  # type: ignore
        available_converters,
        disable_tqdm,
        end_hour,
        irradiance_data,
        logger,
        minigrid,
        number_of_hw_tanks,
        processed_total_hw_load,
        hot_water_pvt_size,
        scenario,
        start_hour,
        temperature_data,
        wind_speed_data,
    )
    logger.debug(
        "Mean hot-water tank temperature: %s",
        np.mean(hot_water_tank_temperature.values)
        if hot_water_tank_temperature is not None
        else "N/A",
    )
    logger.debug(
        "Mean hot-water PV-T electric power per unit: %s",
        np.mean(hot_water_pvt_electric_power_per_unit.values),
    )

    # Calculate electricity-related profiles.
    if total_electric_load is None:
        logger.error(
            "No electric load was supplied to the energy_system.run_simulation method "
            "despite this being necessary for the simulation of energy systems."
        )
        raise InternalError(
            "No electric load was supplied to the energy_system.run_simulation method "
            "despite this being necessary for the simulation of energy systems."
        )
    processed_total_electric_load = pd.DataFrame(
        compute_processed_load_profile(scenario, total_electric_load)[  # type: ignore
            start_hour:end_hour
        ].values
        + clean_water_power_consumed.values
        + hot_water_power_consumed.values
        + thermal_desalination_electric_power_consumed.values
    )

    # Compute the electric input profiles.
    battery_storage_profile: pd.DataFrame
    grid_energy: pd.DataFrame
    kerosene_profile: pd.Series
    load_energy: pd.DataFrame
    renewables_energy: pd.DataFrame
    renewables_energy_map: Dict[RenewableEnergySource, pd.DataFrame] = {
        RenewableEnergySource.PV: pv_power_produced,  # type: ignore
        RenewableEnergySource.CLEAN_WATER_PVT: (
            clean_water_pvt_electric_power_per_unit
        ),
        RenewableEnergySource.HOT_WATER_PVT: hot_water_pvt_electric_power_per_unit,
    }
    renewables_energy_used_directly: pd.DataFrame
    (
        battery_storage_profile,
        grid_energy,
        kerosene_profile,
        load_energy,
        renewables_energy,
        renewables_energy_map,
        renewables_energy_used_directly,
    ) = get_electric_battery_storage_profile(
        clean_water_pvt_size=clean_water_pvt_size,
        grid_profile=grid_profile.iloc[start_hour:end_hour, 0],  # type: ignore
        hot_water_pvt_size=hot_water_pvt_size,
        kerosene_usage=kerosene_usage.iloc[start_hour:end_hour, 0],
        location=location,
        logger=logger,
        minigrid=minigrid,
        processed_total_electric_load=processed_total_electric_load,
        renewables_power_produced=renewables_energy_map,
        scenario=scenario,
        end_hour=end_hour,
        pv_size=pv_size,
        start_hour=start_hour,
    )

    if all(renewables_energy.values == 0):
        logger.warning(
            "%sNo renewable electricity was generated. Continuing with grid and diesel "
            "only.%s",
            BColours.warning,
            BColours.endc,
        )

    # Determine the number of households in the community.
    households = pd.DataFrame(
        population_hourly(location)[  # type: ignore
            simulation.start_year
            * HOURS_PER_YEAR : simulation.end_year
            * HOURS_PER_YEAR
        ].values
    )

    # Initialise battery storage parameters
    maximum_battery_energy_throughput: float = (
        electric_storage_size
        * minigrid.battery.cycle_lifetime
        * minigrid.battery.storage_unit
    )
    initial_battery_storage: float = (
        electric_storage_size
        * minigrid.battery.maximum_charge
        * minigrid.battery.storage_unit
    )
    maximum_battery_storage: float = (
        electric_storage_size
        * minigrid.battery.maximum_charge
        * minigrid.battery.storage_unit
    )
    minimum_battery_storage: float = (
        electric_storage_size
        * minigrid.battery.minimum_charge
        * minigrid.battery.storage_unit
    )
    cumulative_battery_storage_power: float = 0.0
    hourly_battery_storage: Dict[int, float] = {}
    new_hourly_battery_storage: float = 0.0
    battery_health: Dict[int, float] = {}

    # @BenWinchester - Re-order this calculation to use CW and HW power consumed.
    # Initialise tank storage parameters
    (
        hourly_cw_tank_storage,
        initial_cw_tank_storage,
        maximum_cw_tank_storage,
        minimum_cw_tank_storage,
        clean_water_power_consumed_mapping,
    ) = _setup_tank_storage_profiles(
        logger,
        number_of_cw_tanks,
        clean_water_power_consumed,
        ResourceType.CLEAN_WATER,
        scenario,
        minigrid.clean_water_tank,
    )

    (
        hourly_hw_tank_storage,  # pylint: disable=unused-variable
        initial_hw_tank_storage,  # pylint: disable=unused-variable
        maximum_hw_tank_storage,  # pylint: disable=unused-variable
        minimum_hw_tank_storage,  # pylint: disable=unused-variable
        hot_water_power_consumed_mapping,  # pylint: disable=unused-variable
    ) = _setup_tank_storage_profiles(
        logger,
        number_of_hw_tanks,
        hot_water_power_consumed,
        ResourceType.HOT_CLEAN_WATER,
        scenario,
        minigrid.hot_water_tank,
    )

    # Initialise electric desalination paramteters.
    (
        _,
        energy_per_desalinated_litre,
        maximum_water_throughput,
    ) = _calculate_electric_desalination_parameters(
        available_converters, feedwater_sources, logger, scenario
    )

    # Intialise tank accounting parameters
    backup_desalinator_water_supplied: Dict[int, float] = {}
    clean_water_demand_met_by_excess_energy: Dict[int, float] = {}
    clean_water_supplied_by_excess_energy: Dict[int, float] = {}
    conventional_water_supplied: Dict[int, float] = {}
    excess_energy_used_desalinating: Dict[int, float] = {}
    storage_water_supplied: Dict[int, float] = {}
    water_surplus: Dict[int, float] = {}
    # water_deficit: List[float] = []

    # Initialise energy accounting parameters
    energy_surplus: Optional[Dict[int, float]] = {}
    energy_deficit: Optional[Dict[int, float]] = {}
    storage_power_supplied: Dict[int, float] = {}

    # Do not do the itteration if no storage is being used
    if electric_storage_size == 0:
        energy_surplus = None
        energy_deficit = None
    # Carry out the itteration if there is some storage involved in the system.
    else:
        # Begin simulation, iterating over timesteps
        for t in tqdm(
            range(int(battery_storage_profile.size)),
            desc="hourly computation",
            disable=disable_tqdm,
            leave=False,
            unit="hour",
        ):
            # Calculate the electric iteration.
            (
                battery_energy_flow,
                excess_energy,
                new_hourly_battery_storage,
            ) = battery_iteration_step(
                battery_storage_profile,
                hourly_battery_storage,
                initial_battery_storage,
                logger,
                maximum_battery_storage,
                minigrid,
                minimum_battery_storage,
                time_index=t,
            )

            # Calculate the hot-water iteration.

            # Calculate the clean-water iteration.
            excess_energy = cw_tank_iteration_step(
                backup_desalinator_water_supplied,
                clean_water_power_consumed_mapping,
                clean_water_demand_met_by_excess_energy,
                clean_water_supplied_by_excess_energy,
                conventional_cw_source_profiles,
                conventional_water_supplied,
                energy_per_desalinated_litre,
                excess_energy,
                excess_energy_used_desalinating,
                hourly_cw_tank_storage,
                initial_cw_tank_storage,
                logger,
                maximum_battery_storage,
                maximum_cw_tank_storage,
                maximum_water_throughput,
                minigrid,
                minimum_cw_tank_storage,
                new_hourly_battery_storage,
                scenario,
                storage_water_supplied,
                tank_storage_profile,
                time_index=t,
            )

            # Dumped energy and unmet demand
            energy_surplus[t] = excess_energy  # type: ignore
            energy_deficit[t] = max(  # type: ignore
                minimum_battery_storage - new_hourly_battery_storage, 0.0
            )  # Battery too empty

            # Battery capacities and blackouts (if battery is too full or empty)
            new_hourly_battery_storage = min(
                new_hourly_battery_storage, maximum_battery_storage
            )
            new_hourly_battery_storage = max(
                new_hourly_battery_storage, minimum_battery_storage
            )

            # Update hourly_battery_storage
            hourly_battery_storage[t] = new_hourly_battery_storage

            # Update battery health
            (
                cumulative_battery_storage_power,
                maximum_battery_storage,
                minimum_battery_storage,
            ) = _update_battery_health(
                battery_energy_flow,
                battery_health,
                cumulative_battery_storage_power,
                electric_storage_size,
                hourly_battery_storage,
                maximum_battery_energy_throughput,
                minigrid,
                storage_power_supplied,
                time_index=t,
            )
    # Process the various outputs into dataframes.
    # energy_deficit_frame: pd.DataFrame = dict_to_dataframe(energy_deficit)
    if energy_surplus is not None:
        energy_surplus_frame: pd.DataFrame = dict_to_dataframe(energy_surplus, logger)
    else:
        energy_surplus_frame = pd.DataFrame([0] * (end_hour - start_hour))

    if scenario.battery and electric_storage_size > 0:
        battery_health_frame: pd.DataFrame = dict_to_dataframe(battery_health, logger)
        hourly_battery_storage_frame: pd.DataFrame = dict_to_dataframe(
            hourly_battery_storage, logger
        )
        storage_power_supplied_frame: pd.DataFrame = dict_to_dataframe(
            storage_power_supplied, logger
        )
    else:
        battery_health_frame = pd.DataFrame([0] * (end_hour - start_hour))
        hourly_battery_storage_frame = pd.DataFrame([0] * (end_hour - start_hour))
        storage_power_supplied_frame = pd.DataFrame([0] * (end_hour - start_hour))

    if scenario.desalination_scenario is not None:
        backup_desalinator_water_frame: pd.DataFrame = dict_to_dataframe(
            backup_desalinator_water_supplied, logger
        )
        clean_water_demand_met_by_excess_energy_frame: pd.DataFrame = dict_to_dataframe(
            clean_water_demand_met_by_excess_energy, logger
        )
        clean_water_power_consumed = dict_to_dataframe(
            clean_water_power_consumed_mapping, logger
        )
        clean_water_supplied_by_excess_energy_frame: pd.DataFrame = dict_to_dataframe(
            clean_water_supplied_by_excess_energy, logger
        )
        conventional_cw_supplied_frame: pd.DataFrame = dict_to_dataframe(
            conventional_water_supplied, logger
        )
        excess_energy_used_desalinating_frame: pd.DataFrame = dict_to_dataframe(
            excess_energy_used_desalinating, logger
        )
        if hourly_cw_tank_storage is None:
            logger.error(
                "%sNo clean-water tank storage level information was outputted from "
                "the simulation despite non-`None` information being expected.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InternalError(
                "Clean-water tank storage information not computed successfully."
            )
        hourly_cw_tank_storage_frame: pd.DataFrame = dict_to_dataframe(
            hourly_cw_tank_storage, logger
        )
        storage_water_supplied_frame: pd.DataFrame = dict_to_dataframe(
            storage_water_supplied, logger
        )
        water_surplus_frame: pd.DataFrame = dict_to_dataframe(water_surplus, logger)

    # Find unmet energy
    unmet_energy = pd.DataFrame(
        (
            load_energy.values
            + thermal_desalination_electric_power_consumed.values
            + -renewables_energy_used_directly.values
            - grid_energy.values
            - storage_power_supplied_frame.values
        )
    )
    if thermal_desalination_electric_power_consumed is not None:
        unmet_energy = pd.DataFrame(
            (unmet_energy.values + thermal_desalination_electric_power_consumed.values)
        )

    # Determine the times for which the system experienced a blackout.
    blackout_times = ((unmet_energy > 0) * 1).astype(float)

    # Use backup diesel generator if present
    diesel_energy: pd.DataFrame
    diesel_fuel_usage: pd.DataFrame
    diesel_times: pd.DataFrame
    if scenario.diesel_scenario.mode in (DieselMode.BACKUP, DieselMode.BACKUP_UNMET):
        (
            diesel_capacity,
            diesel_energy,
            diesel_fuel_usage,
            diesel_times,
            unmet_energy,
        ) = _calculate_backup_diesel_generator_usage(
            blackout_times,
            minigrid,
            scenario,
            float(np.sum(processed_total_electric_load)),
            unmet_energy,
        )
    elif scenario.diesel_scenario.mode == DieselMode.CYCLE_CHARGING:
        logger.error(
            "%sCycle charing is not currently supported.%s",
            BColours.fail,
            BColours.endc,
        )
        raise InputFileError(
            "scenario inputs", "Cycle charing is not currently supported."
        )
    elif scenario.diesel_scenario.mode == DieselMode.DISABLED:
        diesel_energy = pd.DataFrame([0.0] * int(battery_storage_profile.size))
        diesel_times = pd.DataFrame([0.0] * int(battery_storage_profile.size))
        diesel_fuel_usage = pd.DataFrame([0.0] * int(battery_storage_profile.size))
        diesel_capacity = 0.0
    else:
        logger.error(
            "%sDiesel mode must be specified. Valid modes are %s.%s",
            BColours.fail,
            ", ".join({e.value for e in DieselMode}),
            BColours.endc,
        )
        raise InputFileError(
            "scenario inputs", "Diesel mode must be specified in the scenario file."
        )

    # Find new blackout times, according to when there is unmet energy
    blackout_times = ((unmet_energy > 0) * 1).astype(float)
    # Ensure all unmet energy is calculated correctly, removing any negative values
    unmet_energy = ((unmet_energy > 0) * unmet_energy).abs()  # type: ignore
    # Ensure all unmet clean-water energy is considered.
    clean_water_power_consumed = clean_water_power_consumed.mul(  # type: ignore
        1 - blackout_times
    )
    thermal_desalination_electric_power_consumed = (
        thermal_desalination_electric_power_consumed.mul(  # type: ignore
            1 - blackout_times
        )
    )

    # Find how many kerosene lamps are in use
    kerosene_usage = pd.DataFrame(
        blackout_times.loc[:, 0].mul(kerosene_profile.values)  # type: ignore
    )
    kerosene_mitigation = pd.DataFrame(
        (1 - blackout_times).loc[:, 0].mul(kerosene_profile.values)  # type: ignore
    )

    if scenario.desalination_scenario is not None:
        # Compute the amount of time for which the backup water was able to operate.
        backup_desalinator_water_frame = backup_desalinator_water_frame.mul(  # type: ignore
            1 - blackout_times
        )

        # Find total energy used by the system
        total_energy_used = pd.DataFrame(
            renewables_energy_used_directly.values
            + storage_power_supplied_frame.values
            + grid_energy.values
            + diesel_energy.values
            + excess_energy_used_desalinating_frame.values
        )

        power_used_on_electricity = (
            total_energy_used
            - excess_energy_used_desalinating_frame  # type: ignore
            - clean_water_power_consumed  # type: ignore
            - thermal_desalination_electric_power_consumed  # type: ignore
            - hot_water_power_consumed  # type: ignore
        )

        # Compute the outputs from the itteration stage
        total_cw_supplied: pd.DataFrame = pd.DataFrame(  # type: ignore
            renewable_cw_used_directly.values
            + storage_water_supplied_frame.values
            + backup_desalinator_water_frame.values
            + clean_water_supplied_by_excess_energy_frame.values
            + conventional_cw_supplied_frame.values
        ).mul((1 - blackout_times))

        water_surplus_frame = (  # type: ignore
            (total_cw_supplied - processed_total_cw_load) > 0  # type: ignore
        ) * (
            total_cw_supplied - processed_total_cw_load  # type: ignore
        )

        total_cw_used = total_cw_supplied - water_surplus_frame  # type: ignore

        # Compute when the water demand went unmet.
        # NOTE: This is manually handled to be non-`None`.
        if processed_total_cw_load is None:
            raise InternalError("Processed clean-water load was `None` unexpectedly.")
        unmet_clean_water = pd.DataFrame(
            processed_total_cw_load.values - total_cw_supplied.values  # type: ignore
        )
        unmet_clean_water = unmet_clean_water * (unmet_clean_water > 0)  # type: ignore

        # Convert the PV-T units to kWh.
        if minigrid.pvt_panel is None:
            raise InternalError("Minigrid has no PV-T panel present.")
        clean_water_pvt_electric_power_per_kwh: pd.DataFrame = (
            clean_water_pvt_electric_power_per_unit  # type: ignore
            / minigrid.pvt_panel.pv_unit
        )

        # Find the new clean-water blackout times, according to when there is unmet
        # demand
        clean_water_blackout_times = ((unmet_clean_water > 0) * 1).astype(float)

        # Clean-water system performance outputs
        backup_desalinator_water_frame.columns = pd.Index(
            [ColumnHeader.CLEAN_WATER_FROM_PRIORITISATION.value]
        )
        clean_water_blackout_times.columns = pd.Index(
            [ColumnHeader.CLEAN_WATER_BLACKOUTS.value]
        )
        clean_water_demand_met_by_excess_energy_frame.columns = pd.Index(
            [ColumnHeader.CLEAN_WATER_FROM_EXCESS_ELECTRICITY]  # type: ignore
        )
        clean_water_power_consumed.columns = pd.Index(
            [ColumnHeader.POWER_CONSUMED_BY_DESALINATION.value]
        )
        clean_water_supplied_by_excess_energy_frame.columns = pd.Index(
            [ColumnHeader.CLEAN_WATER_FROM_EXCESS_ELECTRICITY.value]
        )
        conventional_cw_supplied_frame.columns = pd.Index(
            [ColumnHeader.CLEAN_WATER_FROM_CONVENTIONAL_SOURCES.value]
        )
        excess_energy_used_desalinating_frame.columns = pd.Index(
            [ColumnHeader.EXCESS_POWER_CONSUMED_BY_DESALINATION.value]
        )
        hourly_cw_tank_storage_frame.columns = pd.Index(
            [ColumnHeader.CW_TANK_STORAGE_PROFILE.value]
        )
        processed_total_cw_load.columns = pd.Index([ColumnHeader.TOTAL_CW_LOAD.value])
        power_used_on_electricity.columns = pd.Index(
            [ColumnHeader.POWER_CONSUMED_BY_ELECTRIC_DEVICES.value]
        )
        renewable_cw_produced.columns = pd.Index(
            [ColumnHeader.CLEAN_WATER_FROM_RENEWABLES.value]
        )
        renewable_cw_used_directly.columns = pd.Index(
            [ColumnHeader.RENEWABLE_CW_USED_DIRECTLY.value]
        )
        storage_water_supplied_frame.columns = pd.Index(
            [ColumnHeader.CLEAN_WATER_FROM_STORAGE.value]
        )
        total_cw_used.columns = pd.Index([ColumnHeader.TOTAL_CW_CONSUMED.value])
        total_cw_supplied.columns = pd.Index(  # type: ignore
            [ColumnHeader.TOTAL_CW_SUPPLIED.value]
        )
        unmet_clean_water.columns = pd.Index([ColumnHeader.UNMET_CLEAN_WATER.value])
        water_surplus_frame.columns = pd.Index([ColumnHeader.WATER_SURPLUS.value])

        if scenario.pv_t:
            if buffer_tank_temperature is None:
                logger.error(
                    "%sInternal error: buffer tank temperature was None despite buffer "
                    "tanks being present.%s",
                    BColours.fail,
                    BColours.endc,
                )
                raise InternalError(
                    "Buffer tank temperature was expected but was `None`."
                )
            buffer_tank_temperature.columns = pd.Index(
                [ColumnHeader.BUFFER_TANK_TEMPERATURE.value]
            )
            buffer_tank_volume_supplied.columns = pd.Index(
                [ColumnHeader.BUFFER_TANK_OUTPUT.value]
            )
            if clean_water_pvt_collector_output_temperature is None:
                logger.error(
                    "%sInternal error: PV-T output temperature was None despite PV-T "
                    "being present.%s",
                    BColours.fail,
                    BColours.endc,
                )
                raise InternalError(
                    "PV-T output temperature was expected but was `None`."
                )
            clean_water_pvt_collector_output_temperature.columns = pd.Index(
                [ColumnHeader.CW_PVT_OUTPUT_TEMPERATURE.value]
            )
            clean_water_pvt_electric_power_per_kwh.columns = pd.Index(
                [ColumnHeader.CW_PVT_ELECTRICITY_SUPPLIED_PER_KWP.value]
            )
            thermal_desalination_electric_power_consumed.columns = pd.Index(
                [ColumnHeader.POWER_CONSUMED_BY_THERMAL_DESALINATION.value]
            )

    else:
        # Find total energy used by the system
        total_energy_used = pd.DataFrame(
            renewables_energy_used_directly.values
            + storage_power_supplied_frame.values
            + grid_energy.values
            + diesel_energy.values
        )

    if scenario.hot_water_scenario is not None:
        # Process any errors.
        if hot_water_tank_temperature is None:
            raise InternalError("Hot-water tank temperature undefined.")
        if hot_water_pvt_collector_output_temperature is None:
            raise InternalError("Hot-water PV-T output temperature undefined.")
        if minigrid.pvt_panel is None:
            raise InternalError("PV-T panel not defined.")

        # Convert the PV-T units to kWh.
        hot_water_pvt_electric_power_per_kwh: pd.DataFrame = pd.DataFrame(
            hot_water_pvt_electric_power_per_unit  # type: ignore
            / minigrid.pvt_panel.pv_unit
        )
        hot_water_power_consumed.columns = pd.Index(
            [ColumnHeader.POWER_CONSUMED_BY_HOT_WATER.value]
        )

        # Add headers to the columns.
        hot_water_power_consumed.columns = pd.Index(
            [ColumnHeader.POWER_CONSUMED_BY_HOT_WATER.value]
        )
        hot_water_pvt_collector_output_temperature.columns = pd.Index(
            [ColumnHeader.HW_PVT_OUTPUT_TEMPERATURE.value]
        )
        hot_water_pvt_electric_power_per_kwh.columns = pd.Index(
            [ColumnHeader.HW_PVT_ELECTRICITY_SUPPLIED_PER_KWP.value]
        )
        hot_water_tank_temperature.columns = pd.Index(
            [ColumnHeader.HW_TANK_TEMPERATURE.value]
        )
        hot_water_tank_volume_supplied.columns = pd.Index(
            [ColumnHeader.HW_TANK_OUTPUT.value]
        )
        processed_total_hw_load.columns = pd.Index([ColumnHeader.TOTAL_HW_LOAD.value])
        renewable_hw_fraction.columns = pd.Index(
            [ColumnHeader.HW_RENEWABLES_FRACTION.value]
        )

    # System performance outputs
    battery_health_frame.columns = pd.Index([ColumnHeader.BATTERY_HEALTH.value])
    blackout_times.columns = pd.Index([ColumnHeader.BLACKOUTS.value])
    diesel_fuel_usage.columns = pd.Index([ColumnHeader.DIESEL_FUEL_USAGE.value])
    diesel_times.columns = pd.Index([ColumnHeader.DIESEL_GENERATOR_TIMES.value])
    energy_surplus_frame.columns = pd.Index([ColumnHeader.DUMPED_ELECTRICITY.value])
    hourly_battery_storage_frame.columns = pd.Index(
        [ColumnHeader.HOURLY_STORAGE_PROFILE.value]
    )
    households.columns = pd.Index([ColumnHeader.HOUSEHOLDS.value])
    diesel_energy.columns = pd.Index([ColumnHeader.DIESEL_ENERGY_SUPPLIED.value])
    kerosene_mitigation.columns = pd.Index([ColumnHeader.KEROSENE_MITIGATION.value])
    kerosene_usage.columns = pd.Index([ColumnHeader.KEROSENE_LAMPS.value])
    storage_power_supplied_frame.columns = pd.Index(
        [ColumnHeader.ELECTRICITY_FROM_STORAGE.value]
    )
    total_energy_used.columns = pd.Index(
        [ColumnHeader.TOTAL_ELECTRICITY_CONSUMED.value]
    )
    unmet_energy.columns = pd.Index([ColumnHeader.UNMET_ELECTRICITY.value])

    # System details
    system_details = SystemDetails(
        diesel_capacity,
        simulation.end_year,
        {
            converter.name: available_converters.count(converter)
            for converter in available_converters
        },
        clean_water_pvt_size
        * float(
            solar_degradation(minigrid.pvt_panel.lifetime, location.max_years).iloc[
                HOURS_PER_YEAR * (simulation.end_year - simulation.start_year), 0
            ]
        )
        if minigrid.pvt_panel is not None and scenario.desalination_scenario is not None
        else None,
        hot_water_pvt_size
        * float(
            solar_degradation(minigrid.pvt_panel.lifetime, location.max_years).iloc[
                HOURS_PER_YEAR * (simulation.end_year - simulation.start_year), 0
            ]
        )
        if minigrid.pvt_panel is not None and scenario.hot_water_scenario is not None
        else None,
        number_of_buffer_tanks if scenario.desalination_scenario is not None else None,
        number_of_cw_tanks if scenario.desalination_scenario is not None else None,
        number_of_hw_tanks if scenario.hot_water_scenario is not None else None,
        pv_size
        * float(
            solar_degradation(minigrid.pv_panel.lifetime, location.max_years).iloc[
                HOURS_PER_YEAR * (simulation.end_year - simulation.start_year), 0
            ]
        ),
        float(
            electric_storage_size
            * minigrid.battery.storage_unit
            * np.min(battery_health_frame[ColumnHeader.BATTERY_HEALTH.value])
        ),
        {
            converter.name: available_converters.count(converter)
            for converter in available_converters
        },
        clean_water_pvt_size
        if minigrid.pvt_panel is not None and scenario.desalination_scenario is not None
        else None,
        hot_water_pvt_size
        if minigrid.pvt_panel is not None and scenario.hot_water_scenario is not None
        else None,
        number_of_buffer_tanks if scenario.desalination_scenario is not None else None,
        number_of_cw_tanks if scenario.desalination_scenario is not None else None,
        number_of_hw_tanks if scenario.hot_water_scenario is not None else None,
        pv_size,
        float(electric_storage_size * minigrid.battery.storage_unit),
        [source.name for source in required_cw_feedwater_sources]
        if len(required_cw_feedwater_sources) > 0
        else None,
        simulation.start_year,
    )

    # Separate out the various renewable inputs.
    pv_energy = renewables_energy_map[RenewableEnergySource.PV]
    clean_water_pvt_energy = renewables_energy_map[
        RenewableEnergySource.CLEAN_WATER_PVT
    ]
    hot_water_pvt_energy = renewables_energy_map[RenewableEnergySource.HOT_WATER_PVT]
    total_pvt_energy = pd.DataFrame(
        clean_water_pvt_energy.values + hot_water_pvt_energy.values
    )
    total_pvt_energy.columns = pd.Index(
        [ColumnHeader.TOTAL_PVT_ELECTRICITY_SUPPLIED.value]
    )

    # End simulation timer
    timer_end = datetime.datetime.now()
    time_delta = timer_end - timer_start

    # Return all outputs
    system_performance_outputs_list = [
        load_energy,
        total_energy_used,
        unmet_energy,
        blackout_times,
        renewables_energy_used_directly,
        storage_power_supplied_frame,
        grid_energy,
        diesel_energy,
        diesel_times,
        diesel_fuel_usage,
        battery_storage_profile,
        pv_energy,
        renewables_energy,
        hourly_battery_storage_frame,
        energy_surplus_frame,
        battery_health_frame,
        households,
        kerosene_usage,
        kerosene_mitigation,
    ]

    if (
        scenario.desalination_scenario is not None
        or scenario.hot_water_scenario is not None
    ):
        system_performance_outputs_list.append(total_pvt_energy)
    if scenario.desalination_scenario is not None:
        desalination_performance_outputs: List[Optional[pd.DataFrame]] = [
            backup_desalinator_water_frame,
            clean_water_blackout_times,
            clean_water_demand_met_by_excess_energy_frame,
            clean_water_power_consumed,
            clean_water_supplied_by_excess_energy_frame,
            conventional_cw_supplied_frame,
            excess_energy_used_desalinating_frame,
            hourly_cw_tank_storage_frame,
            power_used_on_electricity,
            processed_total_cw_load,
            renewable_cw_produced,
            renewable_cw_used_directly,
            storage_water_supplied_frame,
            total_cw_supplied,
            total_cw_used,
            unmet_clean_water,
            water_surplus_frame,
        ]

        if any(entry is None for entry in desalination_performance_outputs):
            logger.error(
                "%sError saving desalination outputs, simulation returned `None` as "
                "outputs despite non-`None` outputs being expected.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InternalError(
                "Some outputs of the simulation were returned as `None` for "
                "desalination related outputs despite non-`None` outputs being "
                "expected."
            )

        system_performance_outputs_list.extend(
            desalination_performance_outputs  # type: ignore
        )

        if scenario.pv_t:
            clean_water_performance_outputs: List[Optional[pd.DataFrame]] = [
                buffer_tank_temperature,
                buffer_tank_volume_supplied,
                clean_water_pvt_collector_output_temperature,
                clean_water_pvt_electric_power_per_kwh,
                clean_water_pvt_energy,
                thermal_desalination_electric_power_consumed,
            ]

            if any(entry is None for entry in clean_water_performance_outputs):
                logger.error(
                    "%sError saving clean-water outputs, simulation returned `None` as "
                    "outputs despite non-`None` outputs being expected.%s",
                    BColours.fail,
                    BColours.endc,
                )
                raise InternalError(
                    "Some outputs of the simulation were returned as `None` for "
                    "clean-water related outputs despite non-`None` outputs being "
                    "expected."
                )

            system_performance_outputs_list.extend(
                clean_water_performance_outputs  # type: ignore
            )

    if scenario.hot_water_scenario is not None:
        hot_water_performance_outputs: List[Optional[pd.DataFrame]] = [
            hot_water_power_consumed,
            hot_water_pvt_collector_output_temperature,
            hot_water_pvt_electric_power_per_kwh,
            hot_water_pvt_electric_power_per_unit,
            hot_water_pvt_energy,
            hot_water_tank_temperature,
            hot_water_tank_volume_supplied,
            processed_total_hw_load,
            renewable_hw_fraction,
        ]

        if any(entry is None for entry in hot_water_performance_outputs):
            logger.error(
                "%sError saving hot-water outputs, simulation returned `None` as "
                "outputs despite non-`None` outputs being expected.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InternalError(
                "Some outputs of the simulation were returned as `None` for hot-water "
                "related outputs despite non-`None` outputs being expected."
            )

        system_performance_outputs_list.extend(
            hot_water_performance_outputs  # type: ignore
        )

    system_performance_outputs = pd.concat(
        system_performance_outputs_list,
        axis=1,
    )

    return time_delta, system_performance_outputs, system_details


#     def lifetime_simulation(self, optimisation_report):
#         """
#         Simulates a minigrid over its lifetime.

#         Simulates a minigrid system over the course of its lifetime to get the complete
#         technical performance of the system

#         Inputs:
#             - optimisation_report:
#                 Report of outputs from Optimisation().multiple_optimisation_step()

#         Outputs:
#             - lifetime_output:
#                 The lifetime technical performance of the system

#         """
#         # Initialise
#         optimisation_report = optimisation_report.reset_index(drop=True)
#         lifetime_output = pd.DataFrame([])
#         simulation_periods = np.size(optimisation_report, 0)
#         # Iterate over all simulation periods
#         for sim in range(simulation_periods):
#             system_performance_outputs = self.simulation(
#                 start_year=int(optimisation_report["Start year"][sim]),
#                 end_year=int(optimisation_report["End year"][sim]),
#                 pv_size=float(optimisation_report["Initial PV size"][sim]),
#                 electric_storage_size=float(
#                     optimisation_report["Initial storage size"][sim]
#                 ),
#             )
#             lifetime_output = pd.concat(
#                 [lifetime_output, system_performance_outputs[0]], axis=0
#             )
#         return lifetime_output.reset_index(drop=True)

#     #%%
#     # =============================================================================
#     # ENERGY BALANCE FUNCTIONS
#     #       These functions identify the sources and uses of energy in the system,
#     #       such as generation, loads and the overall balance
#     # =============================================================================
#     #%% Energy balance

#     #%% Energy usage



#!/usr/bin/python3
########################################################################################
# __utils__.py - CLOVER Scripts Utility module.                                        #
#                                                                                      #
# Author: Phil Sandwell, Ben Winchester                                                #
# Copyright: Phil Sandwell, 2018                                                       #
# Date created: 24/03/2022                                                             #
#                                                                                      #
# For more information, please email:                                                  #
#   philip.sandwell@gmail.com                                                          #
########################################################################################
"""
__utils__.py - Utility module for CLOVER's Scripts Component.

The utility module contains functionality which is used by the various scripts within
CLOVER designed to improve ease of use.

"""

import argparse
import enum
import os

from contextlib import contextmanager
from logging import Logger
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

import yaml

from ..__utils__ import (
    BColours,
    DEFAULT_SCENARIO,
    InputFileError,
    LOCATIONS_FOLDER_NAME,
    read_yaml,
)
from ..fileparser import INPUTS_DIRECTORY, OPTIMISATION_INPUTS_FILE, OPTIMISATIONS
from ..optimisation.__utils__ import OPTIMISATION_CRITERIA, THRESHOLD_CRITERIA

__all__ = (
    "HpcRunType",
    "HpcOptimisation",
    "HpcSimulation",
    "parse_args_and_hpc_input_file",
    "temporary_optimisations_file",
)

# False:
#   Keyword for parsing no total-load file info.
FALSE: str = "false"

# Type:
#   Keyword used for parsing the type of run taking place.
TYPE: str = "type"


class HpcRunType(enum.Enum):
    """
    Denotes the type of run being carried out on the HPC.

    - SIMULATION:
        Denotes a simulation or multiple simulations.

    - OPTIMISATION:
        Denotes an optimisation or multiple optimisations.

    """

    SIMULATION = "simulation"
    OPTIMISATION = "optimisation"


class InvalidRunError(Exception):
    """Raised when an HPC run is invalid."""

    def __init__(self, msg: str) -> None:
        """
        Instantiate an invalid run error.

        Inputs:
            - msg:
                Message to append.

        """

        super().__init__(f"Invalid HPC run in hpc inputs file: {msg}")


class _BaseHpcRun:  # pylint: disable=too-few-public-methods
    """
    Represents a base run that can be carried out on the HPC.

    .. attribute:: location
        The name of the location.

    .. attribute:: output
        The name of the output file to use.

    .. attribute:: total_load
        Whether a total-load file is being used (True) or not (False).

    .. attribute:: total_load_file
        The name of the total-load file, if relevant.

    .. attribute:: type:
        Whether an optimisation or simulation is being run.

    """

    type: HpcRunType

    def __init__(
        self,
        location: str,
        output: str,
        total_load: bool,
        total_load_file: Optional[str] = None,
    ) -> None:
        """
        Instantiate a :class:`_BaseHpcRun` instance.

        Inputs:
            - location:
                The name of the location to use.
            - output:
                The name of the output file to use.
            - total_load:
                Whether a total-load file should be used.
            - total_load_file:
                If being used, the name of the total-load file.

        """

        self.location = location
        self._output = output
        self.total_load = total_load
        self.total_load_file = total_load_file

    def __init_subclass__(cls, run_type: HpcRunType) -> None:
        """
        Method run when instantiating a :class:`_BaseHpcRun` child.

        Inputs:
            - run_type:
                The type of HPC run being carried out.

        """

        super().__init_subclass__()
        cls.type = run_type

    def __str__(self) -> str:
        """
        Returns a nice-looking `str` representing the run.

        Outputs:
            - A nice-looking `str` representing the run.

        """

        return (
            f"HpcRun(type={self.type}, location={self.location}"
            + f", output={self.output}"
            + f", total_load={self.total_load}"
            + (
                f", total_load_file={self.total_load_file}"
                if self.total_load_file is not None
                else ""
            )
            + ")"
        )

    @property
    def output(self) -> str:
        """
        Return an output name for the class.

        Outputs:
            An output name for the run.

        """

        return self._output


class HpcOptimisation(
    _BaseHpcRun, run_type=HpcRunType.OPTIMISATION
):  # pylint: disable=too-few-public-methods
    """
    Represents an optimisation to be carried out on the HPC.

    """

    def __init__(
        self,
        location: str,
        optimisation: List[Dict[str, Any]],
        optimisation_inputs_data: Dict[str, Any],
        output: str,
        total_load: bool,
        total_load_file: Optional[str] = None,
    ) -> None:
        """
        Instantiate a :class:`HpcOptimisation` instance.

        Inputs:
            - location:
                The name of the location to use.
            - optimisation:
                The optimisation parameters for this particular optimisation, passed in
                as a single entry in a `list` for easy temporary file generation.
            - optimisation_inputs_data:
                The input data for optimisations in general.
            - output:
                The name of the output file to use.
            - total_load:
                Whether a total-load file should be used.
            - total_load_file:
                If being used, the name of the total-load file.

        """

        super().__init__(location, output, total_load, total_load_file)
        self.optimisation: List[Dict[str, Any]] = optimisation
        self.optimisation_inputs_data: Dict[str, Any] = optimisation_inputs_data

    @classmethod
    def from_dict(
        cls,
        input_data: Dict[str, Any],
        logger: Logger,
        optimisation: List[Dict[str, Any]],
        optimisation_inputs_data: Dict[str, Any],
    ) -> Any:
        """
        Creates a :class:`HpcOptimisation` instance based on the inputs provided.

        Inputs:
            - input_data:
                The input information, extracted from the HPC inputs file.
            - logger:
                The logger being used for the run.
            - optimisation:
                The optimisation parameters for this particular optimisation
            - optimisation_inputs_data:
                The input data for optimisations in general.

        Outputs:
            - A :class:`HpcOptimisation` instance based on the input information provided.

        """

        total_load_input: Union[str, bool] = input_data.get("total_load", False)

        if not total_load_input:
            total_load: bool = False
            total_load_file: Optional[str] = None
        elif not isinstance(total_load_input, str):
            logger.error(
                "%sCannot set total-load to be `true`.%s", BColours.fail, BColours.endc
            )
            raise InvalidRunError(
                "Cannot set total-load file to be `true` in HPC input file. Either "
                "`false` or the name of the file can be used."
            )
        else:
            total_load = True
            total_load_file = total_load_input

        return cls(
            input_data["location"],
            optimisation,
            optimisation_inputs_data,
            input_data.get("output", "optimisation_output"),
            total_load,
            total_load_file,
        )

    @property
    def optimisation_inputs(self) -> Dict[str, Any]:
        """
        Assembly optimisation input data for writing to file.

        Outputs:
            - Data ready to be written to a temporary optimisation outputs file.

        """

        if OPTIMISATIONS in self.optimisation_inputs_data:
            return self.optimisation_inputs_data

        self.optimisation_inputs_data[OPTIMISATIONS] = self.optimisation
        return self.optimisation_inputs_data

    @property
    def output(self) -> str:
        """
        Return an output name for the class based on the optimisation criterion etc.

        Outputs:
            An output name for the run.

        """

        output_name: str = f"{self._output}_"

        # Add the optimisation criteria information
        for entry in self.optimisation[0][OPTIMISATION_CRITERIA]:
            output_name += "_".join([f"{key}_{value}" for key, value in entry.items()])

        output_name += "_"

        # Add the threshold-criteria information
        for entry in self.optimisation[0][THRESHOLD_CRITERIA]:
            output_name += "_".join(
                [
                    f"{key}_{str(value).replace('.', '_')}"
                    for key, value in entry.items()
                ]
            )

        return output_name


class HpcSimulation(
    _BaseHpcRun, run_type=HpcRunType.SIMULATION
):  # pylint: disable=too-few-public-methods
    """
    Representst a simulation, or multiple simulations, to be carried out on the HPC.

    .. attribute:: pv_system_size
        The size of the PV system.

    .. attribute:: scenario
        The name of the scenario to use.

    .. attribute:: storage_size
        The size of the storage installed.

    """

    def __init__(
        self,
        location: str,
        output: str,
        pv_system_size: float,
        scenario: str,
        storage_size: float,
        total_load: bool,
        total_load_file: Optional[str] = None,
    ) -> None:
        """
        Instantiate a :class:`HpcSimulation` instance.

        Inputs:
            - location:
                The name of the location to use for the simulation(s).
            - output:
                The name of the output file to use.
            - pv_system_size:
                The size of the PV system installed.
            - scenario:
                The scenario to use for the simulation(s).
            - storage_size:
                The size of the storage system installed.
            - total_load:
                Whether a total-load file should be used for the simulation(s).
            - total_load_file:
                If being used, the name of the total-load file for the simulation(s).

        """

        super().__init__(location, output, total_load, total_load_file)
        self.pv_system_size = pv_system_size
        self.scenario = scenario
        self.storage_size = storage_size

    @classmethod
    def from_dict(cls, input_data: Dict[str, Any], logger: Logger) -> Any:
        """
        Creates a :class:`HpcOptimisation` instance based on the inputs provided.

        Inputs:
            - input_data:
                The input information, extracted from the HPC inputs file.
            - logger:
                The logger being used for the run.

        Outputs:
            - A :class:`HpcOptimisation` instance based on the input information provided.

        """

        total_load_input: str = input_data.get("total_load", False)

        if not total_load_input:
            total_load: bool = False
            total_load_file: Optional[str] = None
        else:
            total_load = True
            total_load_file = total_load_input

        output: str = input_data.get("output", "simulation_outputs")

        try:
            pv_system_size: float = float(input_data["pv_system_size"])
        except KeyError:
            logger.error(
                "%sMissing pv system size information.%s", BColours.fail, BColours.endc
            )
            raise InvalidRunError(
                "Missing pv system size information in input file."
            ) from None
        except ValueError:
            logger.error(
                "%PV system size must be either an integer or float, `str` is not "
                "allowed.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InvalidRunError(
                "PV system size must be of type `float` or `int`."
            ) from None

        scenario: str = input_data.get("scenario", DEFAULT_SCENARIO)

        try:
            storage_size: float = float(input_data["storage_size"])
        except KeyError:
            logger.error(
                "%sMissing storage size information.%s", BColours.fail, BColours.endc
            )
            raise InvalidRunError(
                "Missing storage size information in input file."
            ) from None
        except ValueError:
            logger.error(
                "%sStorage size must be either an integer or float, `str` is not "
                "allowed.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InvalidRunError(
                "Storage size must be of type `float` or `int`."
            ) from None

        return cls(
            input_data["location"],
            output,
            pv_system_size,
            scenario,
            storage_size,
            total_load,
            total_load_file,
        )


def _check_walltime(logger: Logger, walltime: Optional[int]) -> int:
    """
    Checks that the walltime is a valid integer for the HPC.

    Inputs:
        - logger:
            The logger to use for the run.
        - walltime:
            The walltime argument to check.

    Outputs:
        The parsed walltime argument.

    Raises:
        - Exception:
            Raised if the walltime is not valid.

    """
    if walltime is None:
        logger.error(
            "%sWalltime must be specified for HPC runs.%s", BColours.fail, BColours.endc
        )
        raise Exception("Walltime must be specified.")

    if not isinstance(walltime, int):
        logger.error("%sWalltime must be an integer.%s", BColours.fail, BColours.endc)
        raise Exception("Walltime must be an integer.")

    if walltime <= 0:
        logger.error(
            "%sWalltime must be a positive integer.%s", BColours.fail, BColours.endc
        )
        raise Exception("Walltime must be a positive integer.")

    if walltime > 72:
        logger.error(
            "%sMaximum allowed walltime for HPC runs is 72 hours, i.e., 3 days.%s",
            BColours.fail,
            BColours.endc,
        )
        raise Exception("Walltime must be between 1 and 72 hours.")

    return walltime


def _parse_hpc_args(args: List[Any]) -> argparse.Namespace:
    """
    Parses the input arguments to the hpc script to determine the HPC input file.

    Inputs:
        - args:
            Unparsed command-line arguments.

    Outputs:
        - The parsed HPC arguments.

    """

    parser = argparse.ArgumentParser()

    # Runs:
    #   The name of the HPC runs file.
    parser.add_argument("--runs", "-r", type=str, help="The path to the HPC runs file.")

    # Verbose:
    #   Used for generating verbose logs for debugging.
    parser.add_argument(
        "--verbose", "-v", action="store_true", default=False, help=argparse.SUPPRESS
    )

    # Walltime:
    #   Used for specifying the walltime for the runs.
    parser.add_argument(
        "--walltime",
        "-w",
        default=None,
        type=int,
        help="The walltime, in hours, for the HPC runs.",
    )

    return parser.parse_args(args)


def _parse_hpc_input_file(input_filename: str, logger: Logger) -> List[Dict[str, Any]]:
    """
    Parses the HPC input file into a dictionary.

    Inputs:
        - input_filename:
            The name of the input file for the HPC runs.
        - logger:
            The :class:`logging.Logger` to use for the run.

    Outputs:
        - The raw data

    """

    # Open the file and extract the YAML data.
    try:
        filedata = read_yaml(input_filename, logger)
    except FileNotFoundError:
        logger.error(
            "%sHPC input file '%s' could not be found.%s",
            BColours.fail,
            input_filename,
            BColours.endc,
        )
        raise

    # Raise an error if the format of the file is invalid.
    if not isinstance(filedata, list):
        logger.error(
            "%sHPC input file '%s' was not of format `list`.%s",
            BColours.fail,
            input_filename,
            BColours.endc,
        )
        raise InputFileError(
            "hpc run inputs",
            f"The HPC run inputs file '{input_filename}' was not of the format `list`. "
            + "See the user guide/wiki for more information on the format of this file.",
        )

    logger.info("HPC input successfully parsed.")

    return filedata


def _parse_optimisations_to_runs(
    entry: Dict[str, Any], logger: Logger
) -> List[HpcOptimisation]:
    """
    Parses, from a single HPC entry, a series of optimisation runs to carry out.

    Each optimisation entry triggers this script to generate a series of optimisation
    run entries for which a single optimisation will be run, with a dummy input file,
    based on the optimisation parameters within the optimisation_inputs.yaml file.

    Inputs:
        - entry:
            The parsed entry from the file.
        - logger:
            The :class:`logging.Logger` to use for the run.

    """

    # Read the optimisation inputs file.
    optimisation_inputs_file = os.path.join(
        LOCATIONS_FOLDER_NAME,
        entry["location"],
        INPUTS_DIRECTORY,
        OPTIMISATION_INPUTS_FILE,
    )
    optimisation_inputs_data = read_yaml(optimisation_inputs_file, logger)

    # Raise an error if the file was not of type `dict`.
    if not isinstance(optimisation_inputs_data, dict):
        logger.error(
            "Optimisation inputs file data for %s was not of type `dict`. Exiting",
            optimisation_inputs_file,
        )
        raise InputFileError(
            "optimisation inputs", "Optimisation inputs was not of type `dict`."
        )

    # Based on the input optimisations, generate a list of optimisations to carry out.
    optimisations_list = optimisation_inputs_data.pop(OPTIMISATIONS)

    return [
        HpcOptimisation.from_dict(
            entry, logger, [optimisation], optimisation_inputs_data
        )
        for optimisation in optimisations_list
    ]


def _process_hpc_input_file(
    input_filename: str, logger: Logger
) -> List[Union[HpcOptimisation, HpcSimulation]]:
    """
    Parses the HPC input file into a list of runs.

    Inputs:
        - input_filename:
            The name of the input file for the HPC runs.
        - logger:
            The :class:`logging.Logger` to use for the run.

    Outputs:
        - A `list` of the hpc optimisations and simulations to be carried out.

    """

    # Parse the HPC input file.
    filedata = _parse_hpc_input_file(input_filename, logger)

    # Parse this into HPC Optimisations and Simulations.
    runs: List[Union[HpcOptimisation, HpcSimulation]] = []
    for entry in filedata:
        if entry[TYPE] == HpcRunType.OPTIMISATION.value:
            runs.extend(_parse_optimisations_to_runs(entry, logger))
        elif entry[TYPE] == HpcRunType.SIMULATION.value:
            runs.append(HpcSimulation.from_dict(entry, logger))
        else:
            logger.error(
                "%sInvalid HPC run type '%s'. Valid types are %s.%s",
                BColours.fail,
                entry[TYPE],
                ", ".join([str(e.value) for e in HpcRunType]),
                BColours.endc,
            )
            raise InvalidRunError(f"Invalid HPC run type {entry[TYPE]}.")

    # Return this list of HPC optimisations and simulations.
    return runs


def parse_args_and_hpc_input_file(
    args: List[Any], logger: Logger
) -> Tuple[str, List[Union[HpcOptimisation, HpcSimulation]], bool, float]:
    """
    Parses command-line arguments and returns the HPC runs to be carried out.

    Inputs:
        - args:
            Unparsed command-line arguments.

    Outputs:
        - The run file name,
        - A `list` of the hpc optimisations and simulations to be carried out,
        - Whether the logging should be verbose or not,
        - The walltime for the run.

    """

    # Parse the arguments to determine the input file.
    parsed_args = _parse_hpc_args(args)
    logger.info("Command-line arguments successfully parsed.")

    # Check that the walltime is valid.
    walltime = _check_walltime(logger, parsed_args.walltime)

    # Parse the input file to determine the runs to be carried out.
    runs = _process_hpc_input_file(parsed_args.runs, logger)

    # Return these runs along with the filename.
    return parsed_args.runs, runs, parsed_args.verbose, walltime


@contextmanager
def temporary_optimisations_file(
    logger: Logger, run: HpcOptimisation, run_number: int
) -> Iterator[str]:
    """
    Creates a temporary optimisations file.

    Inputs:
        - logger:
            The logger being used for the HPC run.
        - run:
            The run being carried out.
        - run_number:
            The number of the run being carried out.

    Yields:
        - filename:
            The name of the file.

    """

    temp_dirpath: str = os.path.join(
        LOCATIONS_FOLDER_NAME,
        run.location,
        INPUTS_DIRECTORY,
    )
    temp_basepath: str = os.path.join(
        os.path.dirname(OPTIMISATION_INPUTS_FILE),
        f"temp_hpc_{run_number}_{os.path.basename(OPTIMISATION_INPUTS_FILE)}",
    )
    temp_filename: str = os.path.join(temp_dirpath, temp_basepath)

    # Attempt to create the temporary file and yield its name.
    logger.info("Attempting to create temporary file %s", temp_filename)
    try:
        with open(
            temp_filename,
            "w",
        ) as temp_file:
            yaml.dump(run.optimisation_inputs, temp_file)
    except FileNotFoundError as e:
        logger.error("Failed to create temporary optimisations file.")
        print(f"Failed to create temporary optimisations file: {str(e)}")
        raise

    logger.info("Temporary optimisations file successfully created.")
    yield temp_basepath

    # Attempt to remove the temporary file on exit, exit regardless.
    logger.info("Attempting to remove temporary optimsiations file.")
    try:
        os.remove(temp_filename)
    except FileNotFoundError as e:
        logger.error("Could not remove temporary optimisations file: %s", str(e))
        print(f"Could not delete temporary optimisations file: {str(e)}")



#!/usr/bin/python3
########################################################################################
# printer.py - Pretty printing of CLOVER outputs to the console.                       #
#                                                                                      #
# Authors: Phil Sandwell, Ben Winchester                                               #
# Copyright: Phil Sandwell, 2021                                                       #
# Date created: 25/01/2021                                                             #
# License: Open source                                                                 #
########################################################################################
"""
printer.py - The printing module for CLOVER.

In order to print information to the end user in an easily understandable way, CLOVER
utilises this module which exposes functionality for easilly computing and outputting
strings to the end user.

"""

import argparse
from typing import List

from .__utils__ import Scenario
from .optimisation.__utils__ import OptimisationParameters
from .simulation.__utils__ import Minigrid

__all__ = ("generate_optimisation_string", "generate_simulation_string")


def generate_optimisation_string(
    minigrid: Minigrid,
    optimisation_inputs: OptimisationParameters,
    scenario: Scenario,
) -> str:
    """
    Generate and return the optimisation string.

    Inputs:
        - minigrid:
            The :class:`Minigrid` being considered for this run.
        - optimisation_inputs:
            The input parameters for this optimisation.
        - scenario:
            The :class:`__utils__.Scenario` currently being considered.

    Outputs:
        - A single `str` to display to the user when running a CLOVER optimisation.

    """

    optimisation_string_list: List[str] = []

    # Append the PV panel information if relevant.
    if scenario.pv and optimisation_inputs.pv_size is not None:
        optimisation_string_list.append(
            f"- PV resolution of {optimisation_inputs.pv_size.step} units "
            + f"({minigrid.pv_panel.pv_unit} kWp per unit)"
        )

    # Append the battery storage information if relevant.
    if (
        scenario.battery
        and optimisation_inputs.storage_size is not None
        and minigrid.battery is not None
    ):
        optimisation_string_list.append(
            f"- Storage resolution of {optimisation_inputs.storage_size.step} units ("
            + f"{minigrid.battery.storage_unit} kWh per unit)"
        )

    # Append the clean-water information if relevant.
    if scenario.desalination_scenario is not None:
        if (
            optimisation_inputs.cw_pvt_size is not None
            and minigrid.pvt_panel is not None
        ):
            optimisation_string_list.append(
                "- Clean-water PV-T resolution of "
                + f"{optimisation_inputs.cw_pvt_size.step} units "
                + f"({minigrid.pvt_panel.pv_unit} kWp and "
                + f"{minigrid.pvt_panel.thermal_unit} kWth per unit)"
            )
        if (
            optimisation_inputs.clean_water_tanks is not None
            and minigrid.clean_water_tank is not None
        ):
            optimisation_string_list.append(
                "- Clean-water tank resolution of "
                + f"{optimisation_inputs.clean_water_tanks.step} units (1 tank of "
                + f"size {minigrid.clean_water_tank.mass} litres per unit)"
            )

    # Append the hot-water information if relevant.
    if scenario.hot_water_scenario is not None:
        if (
            optimisation_inputs.hw_pvt_size is not None
            and minigrid.pvt_panel is not None
        ):
            optimisation_string_list.append(
                "- Hot-water PV-T resolution of "
                f"{optimisation_inputs.hw_pvt_size.step} units "
                + f"({minigrid.pvt_panel.pv_unit} kWp and "
                + f"{minigrid.pvt_panel.thermal_unit} kWth per unit)"
            )
        if (
            optimisation_inputs.hot_water_tanks is not None
            and minigrid.hot_water_tank is not None
        ):
            optimisation_string_list.append(
                "- Hot-water tank resolution of "
                f"{optimisation_inputs.hot_water_tanks.step} units (1 tank of size "
                + f"{minigrid.hot_water_tank.mass} litres per unit)"
            )

    # Append the converter information.
    optimisation_string_list.extend(
        [
            f"- {converter.name} resolution of {sizing.step} units (1 {converter.name} "
            + f"device of {converter.maximum_output_capacity} max output per unit)"
            for converter, sizing in optimisation_inputs.converter_sizes.items()
        ]
    )

    optimisation_string: str = "\n".join(
        [entry for entry in optimisation_string_list if entry != ""]
    )

    return optimisation_string


def generate_simulation_string(
    minigrid: Minigrid,
    overrided_default_sizes: bool,
    parsed_args: argparse.Namespace,
    scenario: Scenario,
) -> str:
    """
    Generate and return the simulation string.

    Inputs:
        - minigrid:
            The :class:`Minigrid` being considered for this run.
        - overrided_default_sizes:
            Whether the default sizes of various components have been overriden or not.
        - parsed_args:
            The parsed command-line arguments.
        - scenario:
            The :class:`__utils__.Scenario` currently being considered.

    Outputs:
        - A single `str` to display to the user when running a CLOVER simulation.

    """

    simulation_string_list: List[str] = []

    # Append the PV panel information if relevant.
    if scenario.pv and parsed_args.pv_system_size is not None:
        simulation_string_list.append(
            f"- {parsed_args.pv_system_size * minigrid.pv_panel.pv_unit} kWp of PV"
            + (
                (
                    f" ({parsed_args.pv_system_size}x "
                    + f"{minigrid.pv_panel.pv_unit} kWp panels)"
                )
                if overrided_default_sizes
                else ""
            )
        )

    # Append the battery storage information if relevant.
    if scenario.battery and minigrid.battery is not None:
        simulation_string_list.append(
            f"- {parsed_args.storage_size * minigrid.battery.storage_unit} kWh of "
            + "storage"
            + (
                (
                    f" ({parsed_args.storage_size}x "
                    + f"{minigrid.battery.storage_unit} kWh batteries)"
                )
                if overrided_default_sizes
                else ""
            )
        )

    # Append the clean-water information if relevant.
    if scenario.desalination_scenario is not None:
        if (
            parsed_args.clean_water_pvt_system_size is not None
            and minigrid.pvt_panel is not None
        ):
            simulation_string_list.append(
                f"- {parsed_args.clean_water_pvt_system_size} Clean-water PV-T panel "
                + f"units ({minigrid.pvt_panel.pv_unit} kWp PV per unit)\n"
            )
        if minigrid.clean_water_tank is not None:
            simulation_string_list.append(
                f"- {parsed_args.num_clean_water_tanks}x "
                + f"{minigrid.clean_water_tank.mass} litres clean-water storage"
                + "\n"
                if scenario.hot_water_scenario is not None
                else ""
            )

    # Append the hot-water information if relevant.
    if scenario.hot_water_scenario is not None:
        if (
            parsed_args.hot_water_pvt_system_size is not None
            and minigrid.pvt_panel is not None
        ):
            simulation_string_list.append(
                f"- {parsed_args.hot_water_pvt_system_size} Hot-water PV-T panel units "
                + f"({minigrid.pvt_panel.pv_unit} kWp PV per unit)\n"
            )
        simulation_string_list.append(
            f"- {parsed_args.num_hot_water_tanks}x {minigrid.hot_water_tank.mass} "
            + "litres hot-water storage"
            if minigrid.hot_water_tank is not None
            else ""
        )

    simulation_string: str = "\n".join(
        [entry for entry in simulation_string_list if entry != ""]
    )

    return simulation_string



#!/usr/bin/python3
########################################################################################
# __init__.py - Python internals module, no functionality here.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
# Date created: 05/07/2021
# License: Open source
########################################################################################
"""
impact.__init__.py - Init module for the impact component.

The `__init__` module ensures that all of the packages are correctly exposed so that
they can be imported when CLOVER is installed as a package.

"""

from .finance import (
    connections_expenditure,
    COSTS,
    diesel_fuel_expenditure,
    discounted_energy_total,
    discounted_equipment_cost,
    expenditure,
    get_total_equipment_cost,
    ImpactingComponent,
    independent_expenditure,
    total_om,
)
from .ghgs import (
    calculate_connections_ghgs,
    calculate_diesel_fuel_ghgs,
    calculate_grid_ghgs,
    calculate_independent_ghgs,
    calculate_kerosene_ghgs,
    calculate_kerosene_ghgs_mitigated,
    calculate_total_equipment_ghgs,
    calculate_total_om,
    EMISSIONS,
)



#!/usr/bin/python3
########################################################################################
# single_line_simulation.py - Single-line simulation module for optimisation component.#
#                                                                                      #
# Authors: Phil Sandwell                                                               #
# Copyright: Phil Sandwell, 2018                                                       #
# Date created: 01/12/2021                                                             #
# License: Open source                                                                 #
# Most recent update: 01/12/2021                                                       #
#                                                                                      #
# For more information, please email:                                                  #
#     philip.sandwell@gmail.com                                                        #
########################################################################################
"""
single_line_simulation.py - The single-line simulation module for the optimisation comp.

As part of the process of carrying out an optimisation, it is necessary to run
simulations with increased bounds. I.E., should the optimum system for a run involve the
maximum allowed value for any of the parameters, then it may be that an increase in this
parameter will result in a better system.

"""

from logging import Logger
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np  # pylint: disable=import-error
import pandas as pd

from ..simulation import energy_system

from ..__utils__ import (
    BColours,
    Location,
    ProgrammerJudgementFault,
    RenewableEnergySource,
    ResourceType,
    Simulation,
    SystemAppraisal,
)
from ..conversion.conversion import Converter, WaterSource
from ..impact.__utils__ import ImpactingComponent  # pylint: disable=import-error

from .__utils__ import (
    ConverterSize,
    Optimisation,
    get_sufficient_appraisals,
    recursive_iteration,
    SolarSystemSize,
    StorageSystemSize,
    TankSize,
)
from .appraisal import appraise_system


__all__ = ("single_line_simulation",)


def single_line_simulation(  # pylint: disable=too-many-locals, too-many-statements
    conventional_cw_source_profiles: Optional[Dict[WaterSource, pd.DataFrame]],
    converter_sizes: Dict[Converter, ConverterSize],
    cw_pvt_size: SolarSystemSize,
    cw_tanks: TankSize,
    converters: Dict[str, Converter],
    disable_tqdm: bool,
    end_year: int,
    finance_inputs: Dict[str, Any],
    ghg_inputs: Dict[str, Any],
    grid_profile: Optional[pd.DataFrame],
    hw_pvt_size: SolarSystemSize,
    hw_tanks: TankSize,
    irradiance_data: pd.Series,
    kerosene_usage: pd.DataFrame,
    location: Location,
    logger: Logger,
    minigrid: energy_system.Minigrid,
    optimisation: Optimisation,
    potential_system: SystemAppraisal,
    previous_system: Optional[SystemAppraisal],
    pv_system_size: SolarSystemSize,
    start_year: int,
    storage_size: StorageSystemSize,
    temperature_data: pd.Series,
    total_loads: Dict[ResourceType, Optional[pd.DataFrame]],
    total_solar_pv_power_produced: pd.Series,
    wind_speed_data: Optional[pd.Series],
    yearly_electric_load_statistics: pd.DataFrame,
) -> Tuple[
    Dict[Converter, ConverterSize],
    SolarSystemSize,
    TankSize,
    SolarSystemSize,
    TankSize,
    SolarSystemSize,
    StorageSystemSize,
    List[SystemAppraisal],
]:
    """
    Preforms an additional round of simulations.

    If the potential optimum system was found to be an edge case (either maximum PV
    capacity, storage capacity etc.) then this function can be called to carry out
    additional simulation(s).

    Inputs:
        - converter_sizes:
            The largest size of each converter that was simulated.
        - cw_pvt_size:
            The largest clean-water PV-T size that was simulated.
        - cw_pvt_size:
            The largest clean-water tank size that was simulated.
        - disable_tqdm:
            Whether to disable the tqdm progress bars (True) or display them (False).
        - hw_pvt_size:
            The largest hot-water PV-T size that was simulated.
        - hw_pvt_size:
            The largest hot-water tank size that was simulated.
        - pv_size:
            The largest pv size that was simulated.
        - storage_size:
            The largest storage size that was simulated.
        - potential_system:
            The system assumed to be the optimum, before this process
        - previous_system:
            The system that was previously installed

    Outputs:
        - cw_pvt_system_size:
            The clean-water PV-T size of the largest system considered.
        - cw_tanks:
            The clean-water tank size of the largest system considered.
        - hw_pvt_system_size:
            The hot-water PV-T size of the largest system considered.
        - hw_tanks:
            The hot-water tank size of the largest system considered.
        - pv_system_size:
            The pv system size of the largest system considered.
        - storage_size:
            The storage size of the largest system considered.
        - system_appraisals:
            The set of system appraisals considered.

    """

    # Instantiate
    logger.info("Single-line optimisation to be carried out.")
    sufficient_appraisals: List[SystemAppraisal] = []
    system_appraisals: List[SystemAppraisal] = []

    _converter_name_to_converter_mapping = {
        converter.name: converter for converter in converter_sizes
    }
    if potential_system.system_details.initial_converter_sizes is None:
        logger.error(
            "%sInitial converter sizes undefined when calling single-line simulation. "
            "%s",
            BColours.fail,
            BColours.endc,
        )
        raise ProgrammerJudgementFault(
            "single line simulation",
            "Misuse of single-line-simulation function: initial converter sizes must "
            "be defined.",
        )
    potential_converter_sizes = {
        _converter_name_to_converter_mapping[name]: value
        for name, value in potential_system.system_details.initial_converter_sizes.items()
    }
    potential_cw_pvt_size = (
        potential_system.system_details.initial_cw_pvt_size
        if potential_system.system_details.initial_cw_pvt_size is not None
        else 0
    )
    potential_hw_pvt_size = (
        potential_system.system_details.initial_hw_pvt_size
        if potential_system.system_details.initial_hw_pvt_size is not None
        else 0
    )
    potential_num_clean_water_tanks = (
        potential_system.system_details.initial_num_clean_water_tanks
        if potential_system.system_details.initial_num_clean_water_tanks is not None
        else 0
    )
    potential_num_hot_water_tanks = (
        potential_system.system_details.initial_num_hot_water_tanks
        if potential_system.system_details.initial_num_hot_water_tanks is not None
        else 0
    )

    # Determine the static converters based on those that were modelled but were not
    # passed in as part of the maximum system size parameters.
    static_converter_sizes: Dict[Converter, int] = {
        converter: potential_converter_sizes[converter]
        for converter in potential_converter_sizes
        if converter not in converter_sizes
    }

    # Check to see if storage size was an integer number of steps, and increase
    # accordingly.
    if (
        np.ceil(storage_size.max / storage_size.step) * storage_size.step
        == storage_size.max
    ):
        test_storage_size = float(storage_size.max + storage_size.step)
    else:
        test_storage_size = float(
            np.ceil(storage_size.max / storage_size.step) * storage_size.step
        )

    # If storage was maxed out:
    if potential_system.system_details.initial_storage_size == storage_size.max:
        logger.info("Increasing storage size.")

        # Increase and iterate over the various power-generation sizes.
        increased_cw_pvt_system_sizes = sorted(
            range(
                int(cw_pvt_size.min),
                int(np.ceil(cw_pvt_size.max + cw_pvt_size.step)),
                int(cw_pvt_size.step),
            ),
            reverse=True,
        )
        increased_hw_pvt_system_sizes = sorted(
            range(
                int(hw_pvt_size.min),
                int(np.ceil(hw_pvt_size.max + hw_pvt_size.step)),
                int(hw_pvt_size.step),
            ),
            reverse=True,
        )
        increased_pv_system_sizes: List[int] = sorted(
            np.arange(
                pv_system_size.min,
                np.ceil(pv_system_size.max + pv_system_size.step),
                pv_system_size.step,
            ),
            reverse=True,
        )

        # Prep variables for the iteration process.
        component_sizes: Dict[
            Union[Converter, ImpactingComponent, RenewableEnergySource],
            Union[int, float],
        ] = {
            ImpactingComponent.CLEAN_WATER_TANK: potential_num_clean_water_tanks,
            ImpactingComponent.HOT_WATER_TANK: potential_num_hot_water_tanks,
            ImpactingComponent.STORAGE: storage_size.max + storage_size.step,
        }
        parameter_space: List[
            Tuple[
                Union[Converter, ImpactingComponent, RenewableEnergySource],
                str,
                Union[List[int], List[float]],
            ]
        ] = []

        # Add the iterable converter sizes.
        for converter, sizes in converter_sizes.items():
            # Construct the list of available sizes for the given converter.
            simulation_converter_sizes: List[int] = sorted(
                range(
                    int(sizes.min),
                    int(np.ceil(sizes.max + sizes.step)),
                    int(sizes.step),
                ),
                reverse=True,
            )

            if len(simulation_converter_sizes) > 1:
                parameter_space.append(
                    (
                        converter,
                        "simulation"
                        if len(parameter_space) == 0
                        else f"{converter.name} size",
                        simulation_converter_sizes,
                    )
                )
            else:
                component_sizes[converter] = simulation_converter_sizes[0]

        # Add the static converter sizes.
        for converter, size in static_converter_sizes.items():
            component_sizes[converter] = size

        if len(increased_cw_pvt_system_sizes) <= 1:
            component_sizes[
                RenewableEnergySource.CLEAN_WATER_PVT
            ] = potential_cw_pvt_size
        else:
            parameter_space.append(
                (
                    RenewableEnergySource.CLEAN_WATER_PVT,
                    "simulation",
                    increased_cw_pvt_system_sizes,
                )
            )
        if len(increased_hw_pvt_system_sizes) <= 1:
            component_sizes[RenewableEnergySource.HOT_WATER_PVT] = potential_hw_pvt_size
        else:
            parameter_space.append(
                (
                    RenewableEnergySource.HOT_WATER_PVT,
                    "simulation" if len(parameter_space) <= 1 else "hw pv-t size",
                    increased_hw_pvt_system_sizes,
                )
            )
        if len(increased_pv_system_sizes) <= 1:
            component_sizes[
                RenewableEnergySource.PV
            ] = potential_system.system_details.initial_pv_size
        else:
            parameter_space.append(
                (
                    RenewableEnergySource.PV,
                    "simulation" if len(parameter_space) <= 1 else "pv size",
                    increased_pv_system_sizes,
                )
            )

        system_appraisals.extend(
            recursive_iteration(
                conventional_cw_source_profiles,
                disable_tqdm,
                end_year,
                finance_inputs,
                ghg_inputs,
                grid_profile,
                irradiance_data,
                kerosene_usage,
                location,
                logger,
                minigrid,
                optimisation,
                previous_system,
                start_year,
                temperature_data,
                total_loads,
                total_solar_pv_power_produced,
                wind_speed_data,
                yearly_electric_load_statistics,
                component_sizes=component_sizes,
                parameter_space=parameter_space,
                system_appraisals=system_appraisals,
            )
        )

        # If the maximum PV system size isn't a round number of steps, carry out a
        # simulation at this size.
        if (
            np.ceil(pv_system_size.max / pv_system_size.step) * pv_system_size.step
            != pv_system_size.max
        ):
            _, simulation_results, system_details = energy_system.run_simulation(
                int(potential_cw_pvt_size),
                conventional_cw_source_profiles,
                converters,
                disable_tqdm,
                test_storage_size,
                grid_profile,
                int(potential_hw_pvt_size),
                irradiance_data,
                kerosene_usage,
                location,
                logger,
                minigrid,
                potential_num_clean_water_tanks,
                potential_num_hot_water_tanks,
                total_solar_pv_power_produced,
                pv_system_size.max,
                optimisation.scenario,
                Simulation(end_year, start_year),
                temperature_data,
                total_loads,
                wind_speed_data,
            )

            # Appraise the system.
            new_appraisal = appraise_system(
                yearly_electric_load_statistics,
                end_year,
                finance_inputs,
                ghg_inputs,
                location,
                logger,
                previous_system,
                optimisation.scenario,
                simulation_results,
                start_year,
                system_details,
            )

            if get_sufficient_appraisals(optimisation, [new_appraisal]):
                sufficient_appraisals.append(new_appraisal)

        # Update the system details.
        storage_size.max = test_storage_size

    # If the number of clean-water tanks was maxed out:
    if (
        potential_num_clean_water_tanks == cw_tanks.max
        and optimisation.scenario.desalination_scenario is not None
    ):
        logger.info("Increasing number of clean-water tanks.")

        # Increase and iterate over the various clean-water generation sizes.
        increased_cw_pvt_system_sizes = sorted(
            range(
                int(cw_pvt_size.min),
                int(np.ceil(cw_pvt_size.max + cw_pvt_size.step)),
                int(cw_pvt_size.step),
            ),
            reverse=True,
        )
        increased_pv_system_sizes = sorted(
            np.arange(
                pv_system_size.min,
                np.ceil(pv_system_size.max + pv_system_size.step),
                pv_system_size.step,
            ),
            reverse=True,
        )
        increased_storage_sizes = sorted(
            np.arange(
                storage_size.min,
                np.ceil(storage_size.max + storage_size.step),
                storage_size.step,
            ),
            reverse=True,
        )

        # Prep variables for the iteration process.
        component_sizes = {
            ImpactingComponent.CLEAN_WATER_TANK: int(cw_tanks.max + cw_tanks.step),
            RenewableEnergySource.HOT_WATER_PVT: potential_hw_pvt_size,
            ImpactingComponent.HOT_WATER_TANK: potential_num_hot_water_tanks,
        }
        parameter_space = []

        # Add the iterable converter sizes.
        for converter, sizes in converter_sizes.items():
            # Construct the list of available sizes for the given converter.
            simulation_converter_sizes = sorted(
                range(
                    int(sizes.min),
                    int(np.ceil(sizes.max + sizes.step)),
                    int(sizes.step),
                ),
                reverse=True,
            )

            if len(simulation_converter_sizes) > 1:
                parameter_space.append(
                    (
                        converter,
                        "simulation"
                        if len(parameter_space) == 0
                        else f"{converter.name} size",
                        simulation_converter_sizes,
                    )
                )
            else:
                component_sizes[converter] = simulation_converter_sizes[0]

        # Add the static converter sizes.
        for converter, size in static_converter_sizes.items():
            component_sizes[converter] = size

        if len(increased_cw_pvt_system_sizes) <= 1:
            component_sizes[
                RenewableEnergySource.CLEAN_WATER_PVT
            ] = potential_cw_pvt_size
        else:
            parameter_space.append(
                (
                    RenewableEnergySource.CLEAN_WATER_PVT,
                    "simulation",
                    increased_cw_pvt_system_sizes,
                )
            )
        if len(increased_pv_system_sizes) <= 1:
            component_sizes[
                RenewableEnergySource.PV
            ] = potential_system.system_details.initial_pv_size
        else:
            parameter_space.append(
                (
                    RenewableEnergySource.PV,
                    "simulation" if len(parameter_space) <= 1 else "pv size",
                    increased_pv_system_sizes,
                )
            )
        if len(increased_storage_sizes) <= 1:
            component_sizes[
                ImpactingComponent.STORAGE
            ] = potential_system.system_details.initial_storage_size
        else:
            parameter_space.append(
                (
                    ImpactingComponent.STORAGE,
                    "simulation" if len(parameter_space) <= 1 else "storage size",
                    increased_storage_sizes,
                )
            )

        system_appraisals.extend(
            recursive_iteration(
                conventional_cw_source_profiles,
                disable_tqdm,
                end_year,
                finance_inputs,
                ghg_inputs,
                grid_profile,
                irradiance_data,
                kerosene_usage,
                location,
                logger,
                minigrid,
                optimisation,
                previous_system,
                start_year,
                temperature_data,
                total_loads,
                total_solar_pv_power_produced,
                wind_speed_data,
                yearly_electric_load_statistics,
                component_sizes=component_sizes,
                parameter_space=parameter_space,
                system_appraisals=system_appraisals,
            )
        )

    # If the number of hot-water tanks was maxed out:
    if (
        potential_num_hot_water_tanks == hw_tanks.max
        and optimisation.scenario.hot_water_scenario is not None
    ):
        logger.info("Increasing number of hot-water tanks.")

        # Increase and iterate over the various hot-water generation sizes.
        increased_hw_pvt_system_sizes = sorted(
            range(
                int(hw_pvt_size.min),
                int(np.ceil(hw_pvt_size.max + hw_pvt_size.step)),
                int(hw_pvt_size.step),
            ),
            reverse=True,
        )
        increased_pv_system_sizes = sorted(
            np.arange(
                pv_system_size.min,
                np.ceil(pv_system_size.max + pv_system_size.step),
                pv_system_size.step,
            ),
            reverse=True,
        )
        increased_storage_sizes = sorted(
            np.arange(
                storage_size.min,
                np.ceil(storage_size.max + storage_size.step),
                storage_size.step,
            ),
            reverse=True,
        )

        # Prep variables for the iteration process.
        component_sizes = {
            RenewableEnergySource.CLEAN_WATER_PVT: potential_cw_pvt_size,
            ImpactingComponent.CLEAN_WATER_TANK: int(cw_tanks.max + cw_tanks.step),
            ImpactingComponent.HOT_WATER_TANK: potential_num_hot_water_tanks,
        }
        parameter_space = []

        # Add the iterable converter sizes.
        for converter, sizes in converter_sizes.items():
            # Construct the list of available sizes for the given converter.
            simulation_converter_sizes = sorted(
                range(
                    int(sizes.min),
                    int(np.ceil(sizes.max + sizes.step)),
                    int(sizes.step),
                ),
                reverse=True,
            )

            if len(simulation_converter_sizes) > 1:
                parameter_space.append(
                    (
                        converter,
                        "simulation"
                        if len(parameter_space) == 0
                        else f"{converter.name} size",
                        simulation_converter_sizes,
                    )
                )
            else:
                component_sizes[converter] = simulation_converter_sizes[0]

        # Add the static converter sizes.
        for converter, size in static_converter_sizes.items():
            component_sizes[converter] = size

        if len(increased_hw_pvt_system_sizes) <= 1:
            component_sizes[RenewableEnergySource.HOT_WATER_PVT] = potential_hw_pvt_size
        else:
            parameter_space.append(
                (
                    RenewableEnergySource.HOT_WATER_PVT,
                    "simulation",
                    increased_hw_pvt_system_sizes,
                )
            )
        if len(increased_pv_system_sizes) <= 1:
            component_sizes[
                RenewableEnergySource.PV
            ] = potential_system.system_details.initial_pv_size
        else:
            parameter_space.append(
                (
                    RenewableEnergySource.PV,
                    "simulation" if len(parameter_space) <= 1 else "pv size",
                    increased_pv_system_sizes,
                )
            )
        if len(increased_storage_sizes) <= 1:
            component_sizes[
                ImpactingComponent.STORAGE
            ] = potential_system.system_details.initial_storage_size
        else:
            parameter_space.append(
                (
                    ImpactingComponent.STORAGE,
                    "simulation" if len(parameter_space) <= 1 else "storage size",
                    increased_storage_sizes,
                )
            )

        system_appraisals.extend(
            recursive_iteration(
                conventional_cw_source_profiles,
                disable_tqdm,
                end_year,
                finance_inputs,
                ghg_inputs,
                grid_profile,
                irradiance_data,
                kerosene_usage,
                location,
                logger,
                minigrid,
                optimisation,
                previous_system,
                start_year,
                temperature_data,
                total_loads,
                total_solar_pv_power_produced,
                wind_speed_data,
                yearly_electric_load_statistics,
                component_sizes=component_sizes,
                parameter_space=parameter_space,
                system_appraisals=system_appraisals,
            )
        )

    # Check to see if PV size was an integer number of steps, and increase accordingly
    if (
        np.ceil(pv_system_size.max / pv_system_size.step) * pv_system_size.step
        == pv_system_size.max
    ):
        test_pv_size = float(pv_system_size.max + pv_system_size.step)
    else:
        test_pv_size = float(
            np.ceil(pv_system_size.max / pv_system_size.step) * pv_system_size.step
        )

    # If PV was maxed out:
    if potential_system.system_details.initial_pv_size == pv_system_size.max:
        logger.info("Increasing PV size.")

        # Increase and iterate over the various storage sizes and PV-T sizes.
        increased_cw_pvt_system_sizes = sorted(
            range(
                int(cw_pvt_size.min),
                int(np.ceil(cw_pvt_size.max + cw_pvt_size.step)),
                int(cw_pvt_size.step),
            ),
            reverse=True,
        )
        increased_cw_tank_sizes = sorted(
            range(
                int(cw_tanks.min),
                int(np.ceil(cw_tanks.max + cw_tanks.step)),
                int(cw_tanks.step),
            ),
            reverse=True,
        )
        increased_hw_pvt_system_sizes = sorted(
            range(
                int(hw_pvt_size.min),
                int(np.ceil(hw_pvt_size.max + hw_pvt_size.step)),
                int(hw_pvt_size.step),
            ),
            reverse=True,
        )
        increased_hw_tank_sizes = sorted(
            range(
                int(hw_tanks.min),
                int(np.ceil(hw_tanks.max + hw_tanks.step)),
                int(hw_tanks.step),
            ),
            reverse=True,
        )
        increased_storage_sizes = sorted(
            np.arange(
                storage_size.min,
                np.ceil(storage_size.max + storage_size.step),
                storage_size.step,
            ),
            reverse=True,
        )

        # Prep variables for the iteration process.
        component_sizes = {
            RenewableEnergySource.PV: pv_system_size.max + pv_system_size.step,
        }
        parameter_space = []

        # Add the iterable converter sizes.
        for converter, sizes in converter_sizes.items():
            # Construct the list of available sizes for the given converter.
            simulation_converter_sizes = sorted(
                range(
                    int(sizes.min),
                    int(np.ceil(sizes.max + sizes.step)),
                    int(sizes.step),
                ),
                reverse=True,
            )

            if len(simulation_converter_sizes) > 1:
                parameter_space.append(
                    (
                        converter,
                        "simulation"
                        if len(parameter_space) == 0
                        else f"{converter.name} size",
                        simulation_converter_sizes,
                    )
                )
            else:
                component_sizes[converter] = simulation_converter_sizes[0]

        # Add the static converter sizes.
        for converter, size in static_converter_sizes.items():
            component_sizes[converter] = size

        if len(increased_cw_pvt_system_sizes) <= 1:
            component_sizes[
                RenewableEnergySource.CLEAN_WATER_PVT
            ] = potential_cw_pvt_size
        else:
            parameter_space.append(
                (
                    RenewableEnergySource.CLEAN_WATER_PVT,
                    "simulation",
                    increased_cw_pvt_system_sizes,
                )
            )
        if len(increased_cw_tank_sizes) <= 1:
            component_sizes[
                ImpactingComponent.CLEAN_WATER_TANK
            ] = potential_num_clean_water_tanks
        else:
            parameter_space.append(
                (
                    ImpactingComponent.CLEAN_WATER_TANK,
                    "simulation" if len(parameter_space) <= 1 else "cw tanks",
                    increased_cw_tank_sizes,
                )
            )
        if len(increased_hw_pvt_system_sizes) <= 1:
            component_sizes[RenewableEnergySource.HOT_WATER_PVT] = potential_hw_pvt_size
        else:
            parameter_space.append(
                (
                    RenewableEnergySource.HOT_WATER_PVT,
                    "simulation" if len(parameter_space) <= 1 else "hw pv-t size",
                    increased_hw_pvt_system_sizes,
                )
            )
        if len(increased_hw_tank_sizes) <= 1:
            component_sizes[
                ImpactingComponent.HOT_WATER_TANK
            ] = potential_num_hot_water_tanks
        else:
            parameter_space.append(
                (
                    ImpactingComponent.HOT_WATER_TANK,
                    "simulation" if len(parameter_space) <= 1 else "hw tanks",
                    increased_hw_tank_sizes,
                )
            )
        if len(increased_storage_sizes) <= 1:
            component_sizes[
                ImpactingComponent.STORAGE
            ] = potential_system.system_details.initial_storage_size
        else:
            parameter_space.append(
                (
                    ImpactingComponent.STORAGE,
                    "simulation" if len(parameter_space) <= 1 else "storage size",
                    increased_storage_sizes,
                )
            )

        system_appraisals.extend(
            recursive_iteration(
                conventional_cw_source_profiles,
                disable_tqdm,
                end_year,
                finance_inputs,
                ghg_inputs,
                grid_profile,
                irradiance_data,
                kerosene_usage,
                location,
                logger,
                minigrid,
                optimisation,
                previous_system,
                start_year,
                temperature_data,
                total_loads,
                total_solar_pv_power_produced,
                wind_speed_data,
                yearly_electric_load_statistics,
                component_sizes=component_sizes,
                parameter_space=parameter_space,
                system_appraisals=system_appraisals,
            )
        )

        pv_system_size.max = test_pv_size

    # Check to see if clean-water PV-T size was an integer number of steps, and increase
    # accordingly
    if (
        np.ceil(cw_pvt_size.max / cw_pvt_size.step) * cw_pvt_size.step
        == cw_pvt_size.max
    ):
        test_cw_pvt_size = float(cw_pvt_size.max + cw_pvt_size.step)
    else:
        test_cw_pvt_size = float(
            np.ceil(cw_pvt_size.max / cw_pvt_size.step) * cw_pvt_size.step
        )

    # If clean-water PV-T was maxed out:
    if (
        potential_cw_pvt_size == cw_pvt_size.max
        and optimisation.scenario.desalination_scenario
    ):
        logger.info("Increasing clean-water PV size.")

        # Increase and iterate over the various storage sizes and other PV-T sizes.
        increased_cw_tank_sizes = sorted(
            range(
                int(cw_tanks.min),
                int(np.ceil(cw_tanks.max + cw_tanks.step)),
                int(cw_tanks.step),
            ),
            reverse=True,
        )
        increased_hw_pvt_system_sizes = sorted(
            range(
                int(hw_pvt_size.min),
                int(np.ceil(hw_pvt_size.max + hw_pvt_size.step)),
                int(hw_pvt_size.step),
            ),
            reverse=True,
        )
        increased_hw_tank_sizes = sorted(
            range(
                int(hw_tanks.min),
                int(np.ceil(hw_tanks.max + hw_tanks.step)),
                int(hw_tanks.step),
            ),
            reverse=True,
        )
        increased_pv_system_sizes = sorted(
            np.arange(
                pv_system_size.min,
                np.ceil(pv_system_size.max + pv_system_size.step),
                pv_system_size.step,
            ),
            reverse=True,
        )
        increased_storage_sizes = sorted(
            np.arange(
                storage_size.min,
                np.ceil(storage_size.max + storage_size.step),
                storage_size.step,
            ),
            reverse=True,
        )

        # Prep variables for the iteration process.
        component_sizes = {
            RenewableEnergySource.CLEAN_WATER_PVT: int(
                test_cw_pvt_size + cw_pvt_size.step
            )
        }
        parameter_space = []

        # Add the iterable converter sizes.
        for converter, sizes in converter_sizes.items():
            # Construct the list of available sizes for the given converter.
            simulation_converter_sizes = sorted(
                range(
                    int(sizes.min),
                    int(np.ceil(sizes.max + sizes.step)),
                    int(sizes.step),
                ),
                reverse=True,
            )

            if len(simulation_converter_sizes) > 1:
                parameter_space.append(
                    (
                        converter,
                        "simulation"
                        if len(parameter_space) == 0
                        else f"{converter.name} size",
                        simulation_converter_sizes,
                    )
                )
            else:
                component_sizes[converter] = simulation_converter_sizes[0]

        # Add the static converter sizes.
        for converter, size in static_converter_sizes.items():
            component_sizes[converter] = size

        if len(increased_cw_tank_sizes) <= 1:
            component_sizes[
                ImpactingComponent.CLEAN_WATER_TANK
            ] = potential_num_clean_water_tanks
        else:
            parameter_space.append(
                (
                    ImpactingComponent.CLEAN_WATER_TANK,
                    "simulation",
                    increased_cw_tank_sizes,
                )
            )
        if len(increased_hw_pvt_system_sizes) <= 1:
            component_sizes[RenewableEnergySource.HOT_WATER_PVT] = potential_hw_pvt_size
        else:
            parameter_space.append(
                (
                    RenewableEnergySource.HOT_WATER_PVT,
                    "simulation" if len(parameter_space) <= 1 else "hw pv-t size",
                    increased_hw_pvt_system_sizes,
                )
            )
        if len(increased_hw_tank_sizes) <= 1:
            component_sizes[
                ImpactingComponent.HOT_WATER_TANK
            ] = potential_num_hot_water_tanks
        else:
            parameter_space.append(
                (
                    ImpactingComponent.HOT_WATER_TANK,
                    "simulation" if len(parameter_space) <= 1 else "hw tanks",
                    increased_hw_tank_sizes,
                )
            )
        if len(increased_storage_sizes) <= 1:
            component_sizes[
                ImpactingComponent.STORAGE
            ] = potential_system.system_details.initial_storage_size
        else:
            parameter_space.append(
                (
                    ImpactingComponent.STORAGE,
                    "simulation" if len(parameter_space) <= 1 else "storage size",
                    increased_storage_sizes,
                )
            )
        if len(increased_pv_system_sizes) <= 1:
            component_sizes[
                RenewableEnergySource.PV
            ] = potential_system.system_details.initial_pv_size
        else:
            parameter_space.append(
                (
                    RenewableEnergySource.PV,
                    "simulation" if len(parameter_space) <= 1 else "pv size",
                    increased_pv_system_sizes,
                )
            )

        system_appraisals.extend(
            recursive_iteration(
                conventional_cw_source_profiles,
                disable_tqdm,
                end_year,
                finance_inputs,
                ghg_inputs,
                grid_profile,
                irradiance_data,
                kerosene_usage,
                location,
                logger,
                minigrid,
                optimisation,
                previous_system,
                start_year,
                temperature_data,
                total_loads,
                total_solar_pv_power_produced,
                wind_speed_data,
                yearly_electric_load_statistics,
                component_sizes=component_sizes,
                parameter_space=parameter_space,
                system_appraisals=system_appraisals,
            )
        )

        cw_pvt_size.max = test_cw_pvt_size

    # Check to see if hot-water PV-T size was an integer number of steps, and increase
    # accordingly
    if (
        np.ceil(hw_pvt_size.max / hw_pvt_size.step) * hw_pvt_size.step
        == hw_pvt_size.max
    ):
        test_hw_pvt_size = float(hw_pvt_size.max + hw_pvt_size.step)
    else:
        test_hw_pvt_size = float(
            np.ceil(hw_pvt_size.max / hw_pvt_size.step) * hw_pvt_size.step
        )

    # If hot-water PV-T was maxed out:
    if (
        potential_hw_pvt_size == hw_pvt_size.max
        and optimisation.scenario.hot_water_scenario
    ):
        logger.info("Increasing hot-water PV size.")

        # Increase and iterate over the various storage sizes and other PV-T sizes.
        increased_cw_pvt_system_sizes = sorted(
            range(
                int(cw_pvt_size.min),
                int(np.ceil(cw_pvt_size.max + cw_pvt_size.step)),
                int(cw_pvt_size.step),
            ),
            reverse=True,
        )
        increased_cw_tank_sizes = sorted(
            range(
                int(cw_tanks.min),
                int(np.ceil(cw_tanks.max + cw_tanks.step)),
                int(cw_tanks.step),
            ),
            reverse=True,
        )
        increased_hw_tank_sizes = sorted(
            range(
                int(hw_tanks.min),
                int(np.ceil(hw_tanks.max + hw_tanks.step)),
                int(hw_tanks.step),
            ),
            reverse=True,
        )
        increased_pv_system_sizes = sorted(
            np.arange(
                pv_system_size.min,
                np.ceil(pv_system_size.max + pv_system_size.step),
                pv_system_size.step,
            ),
            reverse=True,
        )
        increased_storage_sizes = sorted(
            np.arange(
                storage_size.min,
                np.ceil(storage_size.max + storage_size.step),
                storage_size.step,
            ),
            reverse=True,
        )

        # Prep variables for the iteration process.
        component_sizes = {
            RenewableEnergySource.HOT_WATER_PVT: int(
                test_hw_pvt_size + hw_pvt_size.step
            )
        }
        parameter_space = []

        # Add the iterable converter sizes.
        for converter, sizes in converter_sizes.items():
            # Construct the list of available sizes for the given converter.
            simulation_converter_sizes = sorted(
                range(
                    int(sizes.min),
                    int(np.ceil(sizes.max + sizes.step)),
                    int(sizes.step),
                ),
                reverse=True,
            )

            if len(simulation_converter_sizes) > 1:
                parameter_space.append(
                    (
                        converter,
                        "simulation"
                        if len(parameter_space) == 0
                        else f"{converter.name} size",
                        simulation_converter_sizes,
                    )
                )
            else:
                component_sizes[converter] = simulation_converter_sizes[0]

        # Add the static converter sizes.
        for converter, size in static_converter_sizes.items():
            component_sizes[converter] = size

        if len(increased_cw_pvt_system_sizes) <= 1:
            component_sizes[
                RenewableEnergySource.CLEAN_WATER_PVT
            ] = potential_cw_pvt_size
        else:
            parameter_space.append(
                (
                    RenewableEnergySource.CLEAN_WATER_PVT,
                    "simulation",
                    increased_cw_pvt_system_sizes,
                )
            )
        if len(increased_cw_tank_sizes) <= 1:
            component_sizes[
                ImpactingComponent.CLEAN_WATER_TANK
            ] = potential_num_clean_water_tanks
        else:
            parameter_space.append(
                (
                    ImpactingComponent.CLEAN_WATER_TANK,
                    "simulation",
                    increased_cw_tank_sizes,
                )
            )
        if len(increased_hw_tank_sizes) <= 1:
            component_sizes[
                ImpactingComponent.HOT_WATER_TANK
            ] = potential_num_hot_water_tanks
        else:
            parameter_space.append(
                (
                    ImpactingComponent.HOT_WATER_TANK,
                    "simulation" if len(parameter_space) <= 1 else "hw tanks",
                    increased_hw_tank_sizes,
                )
            )
        if len(increased_storage_sizes) <= 1:
            component_sizes[
                ImpactingComponent.STORAGE
            ] = potential_system.system_details.initial_storage_size
        else:
            parameter_space.append(
                (
                    ImpactingComponent.STORAGE,
                    "simulation" if len(parameter_space) <= 1 else "storage size",
                    increased_storage_sizes,
                )
            )
        if len(increased_pv_system_sizes) <= 1:
            component_sizes[
                RenewableEnergySource.PV
            ] = potential_system.system_details.initial_pv_size
        else:
            parameter_space.append(
                (
                    RenewableEnergySource.PV,
                    "simulation" if len(parameter_space) <= 1 else "pv size",
                    increased_pv_system_sizes,
                )
            )

        system_appraisals.extend(
            recursive_iteration(
                conventional_cw_source_profiles,
                disable_tqdm,
                end_year,
                finance_inputs,
                ghg_inputs,
                grid_profile,
                irradiance_data,
                kerosene_usage,
                location,
                logger,
                minigrid,
                optimisation,
                previous_system,
                start_year,
                temperature_data,
                total_loads,
                total_solar_pv_power_produced,
                wind_speed_data,
                yearly_electric_load_statistics,
                component_sizes=component_sizes,
                parameter_space=parameter_space,
                system_appraisals=system_appraisals,
            )
        )

        hw_pvt_size.max = test_hw_pvt_size

    sufficient_appraisals.extend(
        get_sufficient_appraisals(optimisation, system_appraisals)
    )

    return (
        converter_sizes,
        cw_pvt_size,
        cw_tanks,
        hw_pvt_size,
        hw_tanks,
        pv_system_size,
        storage_size,
        sufficient_appraisals,
    )



#!/usr/bin/python3
########################################################################################
# grid.py - Grid-profile generation module.                                            #
#                                                                                      #
# Author: Phil Sandwell                                                                #
# Copyright: Phil Sandwell, 2018                                                       #
# License: Open source                                                                 #
# Most recent update: 14/07/2021                                                       #
#                                                                                      #
# For more information, please email:                                                  #
#     philip.sandwell@gmail.com                                                        #
########################################################################################
"""
grid.py - The grid-generation module for CLOVER.

This module generates grid-availability profiles for CLOVER.

"""

import os

from logging import Logger
from typing import Dict, Optional

import pandas as pd

from ..__utils__ import BColours, Scenario
from .__utils__ import get_intermittent_supply_status

__all__ = ("get_lifetime_grid_status", "load_grid_profile")


def get_lifetime_grid_status(
    disable_tqdm: bool,
    generation_directory: str,
    grid_times: pd.DataFrame,
    logger: Logger,
    max_years: int,
) -> Dict[str, pd.DataFrame]:
    """
    Calculates, and saves, the grid-availability profiles of all input types.

    Inputs:
        - disable_tqdm:
            Whether to disable tqdm progress bars (True) or display them (False).
        - generation_directory:
            The directory in which auto-generated files should be saved.
        - grid_times:
            Grid inputs information, read from the grid-inputs file.
        - logger:
            The logger to use for the run.
        - max_years:
            The maximum number of years for which the simulation should run.

    Outputs:
        - grid_profiles:
            A dictionary mapping the grid name to the grid profile.

    """

    return get_intermittent_supply_status(
        disable_tqdm, generation_directory, "grid", logger, max_years, grid_times
    )


#     #%%
#     def change_grid_coverage(self, grid_type="bahraich", hours=12):
#         grid_profile = self.grid_times[grid_type]
#         baseline_hours = np.sum(grid_profile)
#         new_profile = pd.DataFrame([0] * 24)
#         for hour in range(24):
#             m = interp1d([0, baseline_hours, 24], [0, grid_profile[hour], 1])
#             new_profile.iloc[hour] = m(hours).round(3)
#         new_profile.columns = [grid_type + "_" + str(hours)]
#         return new_profile

#     def save_grid_coverage(self, grid_type="bahraich", hours=12):
#         new_profile = self.change_grid_coverage(grid_type, hours)
#         new_profile_name = grid_type + "_" + str(hours)
#         output = self.grid_times
#         if new_profile_name in output.columns:
#             output[new_profile_name] = new_profile
#         else:
#             output = pd.concat([output, new_profile], axis=1)
#         output.to_csv(self.generation_filepath + "Grid inputs.csv")


def load_grid_profile(
    auto_generated_files_directory: str, logger: Logger, scenario: Scenario
) -> Optional[pd.DataFrame]:
    """
    Loads the grid profile required for the run.

    Inputs:
        - auto_generated_files_directory:
            The file into which auto-generated files are saved.
        - logger:
            The :class:`logging.Logger` to use for the run.
        - scenario:
            The scenario to use for the run.

    """

    grid_profile: Optional[pd.DataFrame] = None
    if scenario.grid:
        try:
            with open(
                os.path.join(
                    auto_generated_files_directory,
                    "grid",
                    f"{scenario.grid_type}_grid_status.csv",
                ),
                "r",
            ) as f:
                grid_profile = pd.read_csv(
                    f,
                    index_col=0,
                )
        except FileNotFoundError as e:
            logger.error(
                "%sGrid profile file for profile '%s' could not be found: %s%s",
                BColours.fail,
                scenario.grid_type,
                str(e),
                BColours.endc,
            )
            raise

    return grid_profile
